"""Test the Generator Drive model."""

import numpy as np
import pytest

from paraqeet.model.drive_operator import DriveOperator
from paraqeet.signal.envelopes import FlatTopGaussianEnvelope
from paraqeet.signal.iq_mixer import IQMixer

LEN_SIG = 101


@pytest.fixture
def time_samples():
    """Return generated time samples from the given signal length.

    The given signal length `LEN_SIG` is a module level global variable.

    """
    return np.linspace(0, 10e-9, LEN_SIG)


@pytest.fixture
def tone():
    """Generate a sinusoidal tone."""
    tone = FlatTopGaussianEnvelope()
    tone.set_optimizable_parameters(tone.get_parameters())
    return tone


@pytest.fixture
def gen(tone):
    """Return a sinusoidal tone generator object."""
    gen = IQMixer(envelopes=[tone])
    return gen


@pytest.fixture
def drive(gen):
    """Return a generator drive object."""
    drive = DriveOperator(gen, is_longitudinal=False)
    return drive


def test_drive_get_value(drive, time_samples):
    """Test the drive getMatrix method."""
    dim = np.random.randint(2, 10)
    annihilation_op = np.sqrt(np.diag(np.arange(1, dim, dtype=np.float64), k=1))
    driveMatrices = drive.get_value(annihilation_op, time_samples)
    assert driveMatrices.shape == time_samples.shape + (dim, dim)


def test_drive_gradient(tone, drive, time_samples):
    """Test the drive gradient."""
    dim = np.random.randint(2, 10)
    annihilation_op = np.sqrt(np.diag(np.arange(1, dim, dtype=np.float64), k=1))
    grads = drive.get_gradient(annihilation_op, time_samples)
    tone_params = tone.get_parameters()
    assert grads.shape == time_samples.shape + (len(tone_params), dim, dim)


def test_has_parameters(drive):
    assert drive.get_parameters() is not None
    assert len(drive.get_parameters()) >= 0
    assert drive.generator is not None



"""Class definition of the composite Hamiltonian model."""

import jax
import jax.numpy as jnp
import numpy as np

from paraqeet.exceptions import IncompatibleLayersException
from paraqeet.model.coupling import TwoBodyCoupling
from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.quantity import Array, Quantity


class CompositeHamiltonian(DifferentiableHamiltonian):
    """A hamiltonian that consists of subsystems and couplings.

    This class takes care of the tensor products.
    The list of parameters will contain the parameters of all subsystems
    and couplings in the order they were added.

    Parameters
    ----------
    subsystems : list[DifferentiableHamiltonian]
        List of DifferentiableHamiltonians forming the subsystems of a composite system.
    couplings: list[Coupling], optional
        List of couplings between the various subsystems
    """

    _subsystems: list[DifferentiableHamiltonian]
    _couplings: list[TwoBodyCoupling]
    _dimensions: list[int]
    _total_dimension: int

    def __init__(
        self,
        subsystems: list[DifferentiableHamiltonian],
        couplings: list[TwoBodyCoupling] | None = None,
    ):
        super().__init__()
        if couplings is None:
            couplings = []
        self._subsystems = subsystems
        self._couplings = couplings
        self._dimensions = [s.dimension() for s in subsystems]
        self._total_dimension = int(np.prod(self._dimensions))

    def get_parameters(self) -> list[Quantity]:
        """Collect parameters from all subsystems and couplings.

        Parameters
        ----------
        list[Quantity]
            Returns the list of parameters of the system.

        """
        params = []
        for subsystem in self._subsystems:
            params += subsystem.get_parameters()
        for coupling in self._couplings:
            params += coupling.get_parameters()
        return params

    def set_optimizable_parameters(self, params: list[Quantity]) -> None:
        """Set optimizable parameters for the system.

        Forward parameters to the subsystems and couplings.
        All of them should find their own parameters in the list.

        Parameters
        ----------
        params : list[Quantity]
            Input list of parameters to be set.

        """
        for subsystem in self._subsystems:
            subsystem.set_optimizable_parameters(params)
        for coupling in self._couplings:
            coupling.set_optimizable_parameters(params)

    def dimension(self) -> int:
        """Return the dimension of the system.

        Returns
        -------
        int
            Dimension of the system.

        """
        return self._total_dimension

    def get_subsystem_dimensions(self) -> list[int]:
        """Return a list of dimensions of each subsystem in the composite Hamiltonian.

        Returns
        -------
        list[int]
            list of dimension of each subsystem.
        """
        return self._dimensions

    def get_value_at_timestep(self, timestep: float) -> Array:
        """Get matrix representation of the Hamiltonian for a single time point.

        Parameters
        ----------
        timestep: float
            One time step.

        Returns
        -------
        Array
            Hamiltonian of shape [n, n] with 'n' as the Hilbert space
            dimension.

        """
        # Calculate the tensor product of all subsystem matrices
        matrix = jnp.zeros((self._total_dimension, self._total_dimension))
        for n, subsystem in enumerate(self._subsystems):
            sub_matrix = subsystem.get_value_at_timestep(timestep)
            matrix += self._tensor_product_with_identity([sub_matrix], [n])

        for coupling in self._couplings:
            # Create a tensor product where all subsystems
            # except the coupled ones are identity
            indices = [self._subsystems.index(s) for s in coupling.subsystems]
            sub_matrices = coupling.get_couplings()
            for term in sub_matrices:
                matrix += self._tensor_product_with_identity(term, indices)

        return matrix

    def get_gradient_at_timestep(self, time: float) -> Array:
        """Return the gradient of each parameter as an array for one timestamp.

        Collects the gradients from every subsystem and coupling and constructs
        the matrix in the dimension of the composite system.

        Parameters
        ----------
        time: float
            One time point.

        Returns
        -------
        Array
            Gradient at time t.

        """
        gradients = []

        # Take the gradients from all subsystems and plug them into the
        # tensor product with identities
        for one_index, subsystem in enumerate(self._subsystems):
            if not isinstance(subsystem, DifferentiableHamiltonian):
                raise IncompatibleLayersException(f"Expected {subsystem} to provide gradients.")

            # TODO: Fix typing
            # ignoring mypy due to vmap
            sub_gradients = subsystem.get_gradient_at_timestep(jnp.array(time, ndmin=1))  # type: ignore
            for g in sub_gradients:
                if not isinstance(g, np.ndarray | jax.Array):
                    raise IncompatibleLayersException(f"Expected 'Array' got {type(g)} as gradient.")
                gradients.append(self._tensor_product_with_identity([g], [one_index]))

        # Do the same for couplings, except that the tensor product
        # has more than one non-identity component.
        for coupling in self._couplings:
            indices = [self._subsystems.index(s) for s in coupling.subsystems]
            coupling_gradient = coupling.get_coupling_gradients()
            for term in coupling_gradient:
                for g_list in term:
                    grad = self._tensor_product_with_identity(g_list, indices)
                    if grad.size != 0:
                        gradients.append(grad)

        return jnp.array(gradients)

    def _tensor_product_with_identity(self, mat_list: list[Array], n: list[int]) -> Array:
        r"""Put the matrices mat_list into a tensor product at positions `n`.

        All other positions are identity matrices:

        .. math::
            1 \otimes \dots \otimes 1 \otimes mat_list_1 \otimes 1
                \otimes \dots \otimes 1 \otimes mat_list_2 \dots

        The dimensions are assumed to be the same as the subsystems.

        Parameters
        ----------
        mat_list : List[Array]
            List of Matrices for tensor product
        n : list[int]
            List of indices for the each mat_list_i

        Returns
        -------
        Array
            Tensor product of mat_list_i's with I's.

        """
        # Create identity matrices for all subsystems and
        # fill in mat_list at the corresponding indices
        sub_matrices = [jnp.eye(s.dimension()) for s in self._subsystems]
        for i, k in enumerate(n):
            sub_matrices[k] = jnp.array(mat_list[i])

        # Tensor product everything in sub_matrices
        product = jnp.eye(1)
        for m in sub_matrices:
            product = jnp.kron(product, m)

        return product

    def get_collapseops(self) -> list[tuple[Array, Array]]:
        """
        Gather collapse operators from the subsystems and then tensor product them
        with identity to create the collapse operators of the right dimension.
        """
        all_collapse_ops = []
        for n, subsystem in enumerate(self._subsystems):
            rates_and_cols = subsystem.get_collapseops()
            for rate, col_op in rates_and_cols:
                all_collapse_ops.append((rate, self._tensor_product_with_identity([col_op], [n])))
        return all_collapse_ops



"""Class definition for the Scipy optimizer gradient model."""

import jax.numpy as jnp
import numpy as np
from scipy.optimize import minimize

from paraqeet.differentiable import Differentiable
from paraqeet.exceptions import (
    IncompatibleOptimizationMap,
    IncompatibleQuantityException,
)
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.optimizer import OptimizationResult
from paraqeet.optimizers.scipy_optimizer import ScipyOptimizer
from paraqeet.quantity import Array


class ScipyOptimizerGradient(ScipyOptimizer):
    """The Scipy Optimizer gradient model.

    Minimize the outcome of a measurement with the Scipy optimization package.
    """

    _grad_cache: Array  # of shape (n_parameters,)
    _scales: Array

    def __init__(self, measure: NormalizableMeasurement, optimization_map: OptimizationMap) -> None:
        super().__init__(measure, optimization_map)
        params = self._optimization_map.get_all_parameters()
        self._scales = jnp.array([p.get_scale() for p in params]).flatten()

    def optimize(self, times: Array | float) -> OptimizationResult:
        """Optimize via the Scipy optimizer gradient model.

        Performs the actual optimization.

        *Note - If input `times` is a float, then the start time of propagation is implicitly assumed to be zero.
        For an array of times, the first time point is the start time.*

        Returns
        -------
        OptimizationResult
            The result of the optimization.

        """
        self._times = jnp.array([0.0, times]) if isinstance(times, float) else times

        if self._logger:
            self._logger.start()

        self._build_optimizable_index_list()
        self._optimization_map.register_params_with_optimizables()

        init = []
        for qty in self._optimization_map.get_all_parameters():
            init.append(qty.get_reduced_value())

        try:
            result = minimize(
                fun=self._set_parameters_and_measure,
                jac=self._lookup_jac,
                x0=jnp.concatenate(init).flatten(),
                bounds=[(-1, 1)] * self._opt_idxs[-1],
                method=self._method,
                options=self._options,
                callback=self._callback,
            )
        except Exception as e:
            if "_lbfgsb._lbfgsb.setulb: failed to create array from the 7th" + " argument `g`" in str(e):
                raise IncompatibleOptimizationMap(
                    "Number of quantities in optMap differ from number of" + f" gradients computed. \n {e}"
                )
            else:
                raise e

        if self._logger:
            self._logger.stop(str(result))

        return OptimizationResult(
            status=(OptimizationResult.STATUS_SUCCESS if result.success else OptimizationResult.STATUS_FAILED),
            value=result.fun,
            iterations=result.nfev,
            message=result.message,
            raw_result=result,
        )

    def _set_parameters_and_measure(self, values) -> float:
        """Update the parameter values and return measurement result.

        Returns the measurement result including gradient.
        The gradient is stored in a local cache for lookup.
        This tailored for L-BFGS-B or similar algorithms that alternate
        between function and gradient calls.
        Internal callback.

        Parameters
        ----------
        values: Array
            Parameter values for the update.

        Returns
        -------
        Array
            Returns the inverse of the fidelity.

        """
        log = []
        params = self._optimization_map.get_all_parameters()
        for index, val in enumerate(np.split(values, self._opt_idxs[:-1])):  # TODO: Convert to jax
            params[index].set_reduced_value(val)
            log.append(params[index])
        # TODO: what if the self._measure is not Differentiable? -- then this optimizer should not be used.
        if isinstance(self._measure, Differentiable):
            fun, grad = self._measure.get_value_and_gradient(self._times)
            self._grad_cache = grad

            infid = 1.0 - fun
            if self._logger:
                self._logger.log(log, float(infid))
            return float(1 - fun)
        # TODO: which fallback value can be returned here?
        raise IncompatibleQuantityException(
            "Gradient-based optimizer requires a Differentiable measurement; "
            "provided measurement does not implement Differentiable."
        )

    def _lookup_jac(self, values) -> Array:
        """Update the parameter values.

        Return the gradient of a measurement result.
        Internal callback.

        Parameters
        ----------
        values: Array
            Parameter values for the update.

        Returns
        -------
        Array
            Returns the gradient of a measurement result.

        """
        return -1 * self._grad_cache * self._scales



"""Test the quantity object."""

import numpy as np
import numpy.testing as testing
import pytest

from paraqeet.exceptions import IncompatibleQuantityException
from paraqeet.quantity import Quantity


@pytest.fixture
def five():
    """Create a quantity with value '5' and a max value of '15'."""
    return Quantity(5, 0, 15)


@pytest.fixture
def three():
    """Create a quantity with value '3' and a max value of '15'."""
    return Quantity(3, 0, 15)


# constructor
def test_constructor_keeps_value_length():
    """If value, min, and max are arrays of length n, they should not be altered in the constructor."""
    for n in range(1, 100):
        values = np.random.rand(n)
        q = Quantity(values, min_value=0.9 * values, max_value=1.1 * values, unit="")
        assert len(q.get_value()) == n
        assert len(q.get_min_value()) == n
        assert len(q.get_max_value()) == n


def test_updates_bounds_length():
    """If value is of length N while min, and max are scalar, the bounds should be altered into arrays."""
    for n in range(1, 100):
        # In the constructor
        values = np.random.rand(n)
        q = Quantity(values, min_value=0.9 * np.min(values), max_value=1.1 * np.max(values), unit="")
        assert len(q.get_value()) == n

        min_value = q.get_min_value()
        assert len(min_value) == n
        testing.assert_allclose(min_value, min_value[0])

        max_value = q.get_max_value()
        assert len(max_value) == n
        testing.assert_allclose(max_value, max_value[0])

        # In set_limits
        q.set_limits(0.9 * np.min(values), 1.1 * np.max(values))
        min_value = q.get_min_value()
        assert len(min_value) == n
        testing.assert_allclose(min_value, min_value[0])

        max_value = q.get_max_value()
        assert len(max_value) == n
        testing.assert_allclose(max_value, max_value[0])


def test_fails_on_different_lengths(random_from_list):
    """If value, min, and max are arrays of different length, the quantity should raise an exception."""
    for n in range(1, 100):
        values = np.random.rand(n)
        # Test for 20 random shapes other than N if set_value, set_reduced_value, and set_value_and_limits fail
        for _ in range(20):
            min_shape = random_from_list(np.arange(100), n)
            min_values = 0.9 * np.random.rand(min_shape) * np.min(values)
            max_shape = random_from_list(np.arange(100), np.array([n, min_shape]))
            max_values = (1.0 + 0.1 * np.random.rand(max_shape)) * np.max(values)

            # In the constructor
            with pytest.raises(IncompatibleQuantityException):
                Quantity(values, min_value=min_values, max_value=max_values, unit="")

            # In set_limits
            q = Quantity(values, min_value=0.9 * values, max_value=1.1 * values, unit="")
            with pytest.raises(IncompatibleQuantityException):
                q.set_limits(min_values, max_values)
            testing.assert_array_equal(q.get_value(), values)
            testing.assert_almost_equal(q.get_min_value(), 0.9 * values)
            testing.assert_almost_equal(q.get_max_value(), 1.1 * values)

            # In set_value_and_limits
            with pytest.raises(IncompatibleQuantityException):
                q.set_value_and_limits(values, min_values, max_values)
            testing.assert_array_equal(q.get_value(), values)
            testing.assert_almost_equal(q.get_min_value(), 0.9 * values)
            testing.assert_almost_equal(q.get_max_value(), 1.1 * values)


# getter and setter
def test_get(random_quantity_for_values) -> None:
    """Test get_value, get_min_value, and get_max_value.

    For scalar quantities.

    """
    for n in range(1, 100):
        # create a quantity with random values and check if the
        # get functions return the same values
        values = (2 * np.random.random(n) - 1) * np.power(10.0, np.random.randint(-10, 10))
        q = random_quantity_for_values(values)
        testing.assert_allclose(q.get_value(), values)
        testing.assert_array_less(q.get_min_value(), q.get_value())
        testing.assert_array_less(q.get_value(), q.get_max_value())


def test_set(random_quantity_for_values, random_limits_for_quantity) -> None:
    """Test set_value, set_min_value, and set_max_value.

    For scalar quantities.

    """
    for n in range(1, 100):
        # create a random quantity with values that will be overwritten
        values = (2 * np.random.random(n) - 1) * np.power(10.0, np.random.randint(-10, 10))
        q = random_quantity_for_values(values)
        old_min = q.get_min_value()
        old_max = q.get_max_value()

        # set new values within the limits and check that the
        # limits stay unchanged
        new_values = np.random.random(n) * (old_max - old_min) + old_min
        q.set_value(new_values)
        testing.assert_allclose(q.get_value(), new_values)
        testing.assert_allclose(q.get_min_value(), old_min)
        testing.assert_allclose(q.get_max_value(), old_max)

        # set new limits and check that the values stay unchanged
        new_limits = random_limits_for_quantity(new_values)
        q.set_limits(*new_limits)
        testing.assert_allclose(q.get_min_value(), new_limits[0])
        testing.assert_allclose(q.get_max_value(), new_limits[1])
        testing.assert_array_less(q.get_min_value(), q.get_value())
        testing.assert_array_less(q.get_value(), q.get_max_value())


def test_set_fails_on_different_lengths(random_quantity_for_values, random_from_list):
    """Setting a new value should fail if the length of the new value is different than the old one."""
    for n in range(1, 100):
        # create a random quantity with values that shall be overwritten
        values = (2 * np.random.random(n) - 1) * np.power(10.0, np.random.randint(-10, 10))
        q = random_quantity_for_values(values)
        old_value = q.get_value()
        old_min = q.get_min_value()
        old_max = q.get_max_value()

        # Test for 20 random shapes other than N if set_value, set_reduced_value, and set_value_and_limits fail
        for _ in range(20):
            new_shape = random_from_list(np.arange(100), n)
            new_values = np.random.rand(new_shape)
            with pytest.raises(IncompatibleQuantityException):
                q.set_value(new_values)
            with pytest.raises(IncompatibleQuantityException):
                q.set_reduced_value(new_values)
            if n > 1:
                # If the bounds have a shape of N=1, set_value_and_limits would correctly assume that the bounds should
                # be transformed from scalar into array
                with pytest.raises(IncompatibleQuantityException):
                    q.set_value_and_limits(new_values, old_min, old_max)

        # Make sure that the quantity did not change
        testing.assert_array_equal(q.get_value(), old_value)
        testing.assert_array_equal(q.get_min_value(), old_min)
        testing.assert_array_equal(q.get_max_value(), old_max)


def test_get_item(random_quantity_for_values) -> None:
    """Test get item for scalar quantities."""
    for n in range(1, 100):
        values = (2 * np.random.random(n) - 1) * np.power(10.0, np.random.randint(-10, 10))
        q = random_quantity_for_values(values)
        for i in range(len(q)):
            testing.assert_almost_equal(q[i], values[i])


def test_len(random_quantity_for_values) -> None:
    """Test the length value for scalar quantities."""
    for n in range(1, 100):
        values = (2 * np.random.random(n) - 1) * np.power(10.0, np.random.randint(-10, 10))
        q = random_quantity_for_values(values)
        testing.assert_almost_equal(len(q), n)


# conversion
def test_float(random_quantity_for_values) -> None:
    """Test float conversion for scalar quantities."""
    for _ in range(100):
        value = (2 * np.random.random(1) - 1) * np.power(10.0, np.random.randint(-10, 10))
        q = random_quantity_for_values(value)
        testing.assert_almost_equal(float(q), value)


def test_to_array(random_quantity_for_values) -> None:
    """Test array conversion for scalar quantities."""
    for n in range(1, 100):
        values = (2 * np.random.random(n) - 1) * np.power(10.0, np.random.randint(-10, 10))
        q = random_quantity_for_values(values)
        testing.assert_array_almost_equal(np.asarray(q.get_value()), values)


# comparison
def test_comparisons(random_quantity_for_values) -> None:
    """Tests __lt__, __le__, __gt__, and __ge__ for scalar quantities."""
    for i in range(1, 100):
        # create a quantity with random values and check if the
        # get functions return the same values
        value = (2 * np.random.random() - 1) * np.power(10.0, np.random.randint(-10, 10))
        q1 = random_quantity_for_values(value)
        q2 = random_quantity_for_values(2 * np.abs(value))

        assert q2 > q1
        assert q2 >= q1
        assert q1 < q2
        assert q1 <= q2
        assert q1 <= q1
        assert q1 >= q1


def test_equality(random_quantity_for_values) -> None:
    """Tests __eq__ and __ne__ for scalar quantities."""
    for i in range(1, 100):
        # create a quantity with random values and check if the
        # get functions return the same values
        value = (2 * np.random.random() - 1) * np.power(10.0, np.random.randint(-10, 10))
        q1 = random_quantity_for_values(value)
        q2 = random_quantity_for_values(1.2 * np.abs(value))

        assert q1 != q2
        assert q2 != q1
        assert q1 == q1
        assert q2 == q2
        testing.assert_almost_equal(q1, value)
        testing.assert_almost_equal(q1, q1.get_value())


def test_equality_by_id(random_quantity):
    """Test __eq__ by id."""
    q1 = random_quantity(1)
    q2 = Quantity(q1.get_value(), q1.get_min_value(), q1.get_max_value(), q1.get_unit())
    assert q1 == q2
    assert q1 is not q2
    assert q2 is not q1
    assert q1 is q1
    assert q2 is q2


def test_no_input():
    """Trying to instantiate without any parameters."""
    with pytest.raises(Exception):
        Quantity(np.random.random())
    with pytest.raises(Exception):
        Quantity()
    with pytest.raises(Exception):
        Quantity(np.random.random(), min_value=np.random.random(), max_value=None)
    with pytest.raises(Exception):
        Quantity(np.random.random(), min_value=None, max_value=np.random.random())


def test_out_of_bounds():
    """Test out of bounds for scalar quantities."""
    num = Quantity(5, 3, 6)
    with pytest.raises(Exception):
        num.set_value(7)


def test_range_too_small():
    """Test if a too small range raises an exception."""
    r = np.random.random()

    with pytest.raises(Exception):
        Quantity(r, r, r)
    with pytest.raises(Exception):
        Quantity(r + 1e-10, r, r + 1e-20)

    num = Quantity(r, r - 1, r + 1)
    with pytest.raises(Exception):
        num.set_limits(r, r)
    with pytest.raises(Exception):
        num.set_value_and_limits(r, r, r)


def test_arithmetic(five, three):
    """Test arithmetic special methods for scalar quantities."""
    testing.assert_almost_equal(five + three, 8)
    testing.assert_almost_equal(5 + three, 8)
    testing.assert_almost_equal(five - three, 2)
    testing.assert_almost_equal(five * three, 15)
    testing.assert_almost_equal(five / three, 5.0 / 3)
    testing.assert_almost_equal(5 / three, 5.0 / 3)
    testing.assert_almost_equal(five % 3, 5.0 % 3)
    testing.assert_almost_equal(five * 3, 15)
    testing.assert_almost_equal(5 * three, 15)
    testing.assert_almost_equal(five**1, 5)
    testing.assert_almost_equal(1**three, 1)


def test_str(five):
    """Test string conversion special methods for scalar quantities."""
    volts = Quantity(0.005, 0, 1, unit="V")
    assert str(volts) == "5 mV"
    resist = Quantity(2100, 0, 2500, unit="Ohm")
    assert str(resist) == "2.1 KOhm"
    amps = Quantity(125e6 * 2 * np.pi, 100e6, 1e9, unit="Hz", two_pi=True)
    assert str(amps) == "125 MHz x 2pi"
    bits = Quantity(np.array([256, 512]), 8, 1024, unit="Bits")
    assert str(bits) == "[256 Bits, 512 Bits]"


def test_is_scalar_or_vector(random_quantity):
    """Test whether quantity objects are scalars or vectors."""
    for i in range(20):
        q = random_quantity(1)
        assert q.is_scalar()
        assert not q.is_vector()

    for dim in range(2, 100):
        for i in range(20):
            q = random_quantity(dim)
            assert q.is_vector()
            assert not q.is_scalar()

            v = __generate_random_matrix(dim)
            q2 = Quantity(v, v - 1, v + 1)
            assert not q2.is_vector()
            assert not q2.is_scalar()


def __generate_random_matrix(n: int) -> np.ndarray:
    """Generate a random matrix of size `n` by `n`.

    Parameters
    ----------
    n: int
        Dimension of the matrix.

    Returns
    -------
    Array
        Returns a randomly generated `N` by `N` matrix.

    """
    magnitude: float = np.power(10.0, np.random.randint(-10, 10))
    return (2 * np.random.random((n, n)) - 1) * magnitude


def test_relations(three):
    """Test relation special methods for scalar quantities."""
    relation = Quantity.relational(three, lambda x: 2 * x)
    assert relation.dependent
    assert relation.get_value() == 6
    assert relation.get_min_value() == 0
    assert relation.get_max_value() == 15
    assert relation.get_unit() == ""
    assert relation.get_name() == "relation_of_"

    assert three.dependents == [relation]
    assert relation.dependencies == [three]

    three.set_value(4)
    assert relation.get_value() == 8
    with pytest.raises(ValueError):
        relation.set_value(7)

    copy = Quantity.relational_copy(relation)
    assert copy.dependent
    assert copy.get_value() == 8
    assert copy.get_min_value() == 0
    assert copy.get_max_value() == 15
    assert copy.get_unit() == ""
    assert copy.get_name() == "relation_of_"

    unit1 = Quantity(1, np.array(0), np.array(10), "Hz", "one")
    unit2 = Quantity(1, np.array(0), np.array(10), "s", "two")
    with pytest.raises(ValueError):
        _ = Quantity.relational([unit1, unit2], lambda x, y: x + y)
    with pytest.raises(ValueError):
        unit1.add_relation(unit2, lambda x: x)


def test_persistence(random_quantity):
    for n in range(1, 10):
        for _ in range(20):
            q = random_quantity(n)
            data = q.to_dict()
            q2 = Quantity(0, -1, 1)
            q2.from_dict(data)

            # assert q == q2
            assert q.get_name() == q2.get_name()
            assert q.get_unit() == q2.get_unit()
            testing.assert_allclose(q.get_min_value(), q2.get_min_value())
            testing.assert_allclose(q.get_max_value(), q2.get_max_value())
            testing.assert_allclose(q.get_reduced_value(), q2.get_reduced_value())
            if n == 1:
                assert q2.is_scalar()
            else:
                assert q2.is_vector()



"""Class definition of the Logger model."""

from abc import ABC
from datetime import datetime

from paraqeet.quantity import Quantity


class Logger(ABC):
    """Abstract base class that can be used as a callback in the optimizer."""

    _start_time: datetime
    _stop_time: datetime
    _counter: int

    def start(self):
        """Start logging and set starting values to the run parameters."""
        self._start_time = datetime.now()
        self._counter = 0

    def log(self, params: list[Quantity], infidelity: float):
        """Template function to direct what happens at each log call.

        Parameters
        ----------
        params: list[Quantity]
            List of parameters to be logged.
        infidelity: float
            Goal value to be logged.

        """
        self._counter += 1

    def stop(self, result_message: str | None = None):
        """Template function to stop logging and set end of log parameters.

        Parameters
        ----------
        result_message : str | None = None
            The message that the user wants to write at the end of the log file.

        """
        self._stop_time = datetime.now()



"""Testing the GRAPE optimization of a TLS system."""

import numpy as np
import pytest

from paraqeet.measurement.state_transfer_fidelity import (
    StateTransferFidelityGRAPE,
)
from paraqeet.model.closed_system import ClosedSystem
from paraqeet.model.open_system import OpenSystem
from paraqeet.model.qubit import Qubit
from paraqeet.model.rotating_frame_drive import RotatingFrameDrive
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.scipy_optimizer_gradient import ScipyOptimizerGradient
from paraqeet.propagation.scipy_expm_grape import ScipyExpmGRAPE
from paraqeet.propagation.vern7_grape import Vern7GRAPE
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import GaussEnvelope
from paraqeet.signal.pwc_generator import PWCGenerator

T_FINAL = 20e-9
FREQ = 1e6
TLIST = np.linspace(0, T_FINAL, 21)
T1 = Quantity(10e-6, 1e-9, 100e-6)
TEMP = Quantity(10e-3, 1e-3, 50e-3)
T2STAR = Quantity(10e-6, 1e-9, 100e-6)


@pytest.fixture
def tone():
    tone = GaussEnvelope(amplitude=Quantity(np.pi / T_FINAL / 3, -np.pi / T_FINAL, np.pi / T_FINAL))
    tone.t_final.set_value(T_FINAL)
    return tone


@pytest.fixture
def pwc_gen(tone):
    gen = PWCGenerator(envelopes=[tone], tlist=TLIST)
    gen.multiply_flat_top = True
    gen.max_amplitude = 2 * 1e8
    return gen


@pytest.fixture(scope="function", params=["open_system", "closed_system"])
def model(pwc_gen, request):
    drive = RotatingFrameDrive(pwc_gen)
    controlled_qubit = Qubit(Quantity(FREQ, FREQ / 4, FREQ), drives=[drive], t1=T1, temp=TEMP, t2star=T2STAR)
    if request.param == "open_system":
        model = OpenSystem(controlled_qubit)
    elif request.param == "closed_system":
        model = ClosedSystem(controlled_qubit)
    return model


@pytest.fixture(scope="function", params=["expm", "ode"])
def states(model, request):
    """Compare the overlap of the initial and final state."""
    if request.param == "expm":
        prop_method = ScipyExpmGRAPE(model=model, resolution=2e9)
        if prop_method.is_open:
            pytest.skip("Currently, ScipyExpmGRAPE is not implemented for open system.")
    elif request.param == "ode":
        model.ode_propagation = True
        prop_method = Vern7GRAPE(model=model, resolution=10e9)

    init = np.array([[1.0], [0.0]])
    target = np.array([[0.0], [1]])

    if prop_method.is_open:
        init = np.matmul(init, init.T)
        target = np.matmul(target, target.T)

    prop_method.set_initial_state(init)
    prop_method.set_target_state(target)

    return StateTransferFidelityGRAPE(
        propagation=prop_method,
        initial_state=init,
        target_state=target,
    )


@pytest.fixture
def opt_map(pwc_gen):
    """Create an optimization map."""
    optmap = OptimizationMap()
    optmap.add(pwc_gen)
    return optmap


@pytest.fixture
def opt(states, opt_map):
    """Create a scipy optimizer gradient object over states."""
    return ScipyOptimizerGradient(measure=states, optimization_map=opt_map)


def test_optim_grape(opt) -> None:
    """Check that the optimization goes below threshold."""
    res = opt.optimize(times=TLIST)
    assert res.value < 1e-2



"""Test the weighted sum goal."""

import itertools

import numpy as np
import pytest

from paraqeet.differentiable import Differentiable
from paraqeet.exceptions import ConfigurationException
from paraqeet.measurement.unitary_fidelity import UnitaryFidelity
from paraqeet.measurement.weighted_sum_goal import WeightedSumGoal
from paraqeet.quantity import Array
from tests.measurement.constant_measurement import ConstantMeasurement
from tests.propagation.identity_propagation import IdentityPropagation


@pytest.fixture
def random_meas(random_unitary_matrix):
    meas_list = []
    num_meas = 10
    for _ in range(num_meas):
        dim = 4
        gate = random_unitary_matrix(dim)
        propagation = IdentityPropagation()
        propagation.set_initial_state(gate)
        cz = np.identity(4)
        cz[-1, -1] = -1.0
        meas_list.append(UnitaryFidelity(propagation, cz))
    return meas_list


def test_weighted_sum_goal(random_meas):
    """Test the weighted sum goal function."""
    weights = np.random.random(len(random_meas))
    weights /= sum(weights)
    goal = WeightedSumGoal(measurements=random_meas, weights=weights)
    times = np.array([1.0])
    assert goal.measure(times=times) >= 0
    # Rounding errors might cause the value to be slightly larger than 1
    assert 0 <= np.round(goal.calculate_normalized_scalar(times=times), 8) <= 1


def test_weighted_sum_goal_options(random_meas):
    """Test that the key error on the options is caught correctly"""
    weights = np.random.random(len(random_meas))
    weights /= sum(weights)
    meas_bool = [True for _ in range(len(random_meas))]
    options = {"value": 1.0, "meas_bool": meas_bool}
    with pytest.raises(KeyError, match=r".* weight and meas_bool ."):
        WeightedSumGoal(measurements=random_meas, weights=weights, sum_of_squares_options=options)
    options = {"weight": 1.0, "meas_list": meas_bool}
    with pytest.raises(KeyError, match=r".* weight and meas_bool ."):
        WeightedSumGoal(measurements=random_meas, weights=weights, sum_of_squares_options=options)


def test_weighted_sum_goal_sum_of_squares(random_meas):
    """Test the correct calculation of the weighted sum goal function"""
    weights = np.random.random(len(random_meas))
    weight_sum_of_squares = -0.05
    total_sum_of_weights = np.sum(weights) + weight_sum_of_squares
    weights = weights / total_sum_of_weights
    weight_sum_of_squares = weight_sum_of_squares / total_sum_of_weights
    meas_list_bool = [True for _ in range(len(random_meas))]
    meas_list_bool[-2] = False
    meas_list_bool[-1] = False
    sum_of_squares_options = {"weight": weight_sum_of_squares, "meas_bool": meas_list_bool}
    goal = WeightedSumGoal(measurements=random_meas, weights=weights, sum_of_squares_options=sum_of_squares_options)
    times = np.array([1.0])
    values = [m.measure(times=times) for m in random_meas]
    sum_meas: Array | float = 0.0
    for ii, w in enumerate(weights):
        sum_meas += w * values[ii]
    sum_square_diff: Array | float = 0.0
    values_in_sum_of_squares = [val for val, flag in zip(values, sum_of_squares_options["meas_bool"]) if flag]
    for meas_a, meas_b in itertools.combinations(values_in_sum_of_squares, 2):
        sum_square_diff += (meas_a - meas_b) ** 2
    sum_meas += sum_of_squares_options["weight"] * sum_square_diff
    assert np.abs(goal.measure(times=times) - sum_meas) <= 1e-8


def test_weighted_sum_goal_sum_of_squares_gradient():
    """Test the gradient only for the sum of squares
    part of the cost function
    """
    num_meas = 10
    weights = np.zeros(num_meas)
    weight_sum_of_squares = 1.0
    meas_list = []
    for _ in range(num_meas):
        value = np.random.rand()
        meas_list.append(ConstantMeasurement(propagation=IdentityPropagation(), value=value))
    meas_list_bool = [True for _ in range(num_meas)]
    sum_of_squares_options = {"weight": weight_sum_of_squares, "meas_bool": meas_list_bool}
    goal = WeightedSumGoal(measurements=meas_list, weights=weights, sum_of_squares_options=sum_of_squares_options)
    times = np.array([1.0])
    _, grad = goal.get_value_and_gradient(times=times)
    values_and_gradients = [m.get_value_and_gradient(times=times) for m in meas_list if isinstance(m, Differentiable)]
    sum_grads = np.zeros_like(values_and_gradients[0][1])
    for ii, w in enumerate(weights):
        sum_grads += w * values_and_gradients[ii][1]
    values_and_gradients_in_sum_of_squares = [
        val_grad for val_grad, flag in zip(values_and_gradients, sum_of_squares_options["meas_bool"]) if flag
    ]
    grads_diff = np.zeros_like(values_and_gradients_in_sum_of_squares[0][1])

    for meas_a, meas_b in itertools.combinations(values_and_gradients_in_sum_of_squares, 2):
        grads_diff += 2 * (meas_a[0] - meas_b[0]) * (meas_a[1] - meas_b[1])
    sum_grads += sum_of_squares_options["weight"] * grads_diff
    assert np.abs(np.sum(sum_grads - grad)) < 1e-9


def test_weighted_sum_goal_mismatched_weights():
    """Test the mismatched weights from a weighted sum goal.

    Raises
    ------
    paraqeet.Exceptions.ConfigurationException
        Raise an exception with the weighted sum goal with weights.

    """
    with pytest.raises(ConfigurationException):
        WeightedSumGoal(measurements=[], weights=[0.2, 0.3, 0.5])


def test_weighted_sum_goal_weights_not_normalized():
    """Test the not normalized weighted sum goal function.

    Raises
    ------
    paraqeet.Exceptions.ConfigurationException
        Raise an exception with the weighted sum goal with weights.

    """
    with pytest.raises(UserWarning):
        WeightedSumGoal(measurements=[None, None, None], weights=[0.4, 0.3, 0.5])



"""Class definition of the Scipy optimizer model."""

from collections.abc import Callable

import jax.numpy as jnp
import numpy as np
from scipy.optimize import minimize

from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.optimizer import OptimizationResult, Optimizer
from paraqeet.quantity import Array


class ScipyOptimizer(Optimizer):
    """Minimize the outcome of a measuremnt with the scipy optimization package.

    Parameters
    ----------
    measure: Measurement
        Implementation of the Measurement class that measures the observable
        to be maximised.
    optimization_map: OptimizationMap
        An optimization map containing all parameters that can be optimized.

    """

    _measure: NormalizableMeasurement
    _opt_idxs: list[int]
    _options: dict
    _method: str
    _callback: Callable | None

    def __init__(self, measure: NormalizableMeasurement, optimization_map: OptimizationMap) -> None:
        super().__init__(measure, optimization_map)
        self._options = {"disp": True}
        self._method = "L-BFGS-B"
        self._callback = None

    @property
    def method(self) -> str:
        """Returns the currently selected optimization method."""
        return self._method

    @method.setter
    def method(self, method: str) -> None:
        """Select method from scipy.optimize.minimize.

        See Also
        --------
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html

        Parameters
        ----------
        method: str
            Type of solver, specified by string value.

        """
        self._method = method

    def set_options(self, opts: dict):
        """Set the options for the system."""
        self._options.update(opts)

    def update_option(self, key, val):
        """Updates one option for the system."""
        self._options[key] = val

    @property
    def callback(self) -> Callable | None:
        """Returns the callback function."""
        return self._callback

    @callback.setter
    def callback(self, cbfun: Callable) -> None:
        """Set the callback function for the optimizer.

        Parameters
        ----------
        Callable
            The function to be set as the callback.

        """
        self._callback = cbfun

    def optimize(self, times: Array | float) -> OptimizationResult:
        """Optimize the system via the Scipy optimizer.

        Performs the actual optimization.

        Since the search parameters are dimensionless and bound by [-1, 1], we set the bounds of the scipy minimize
        module to -1, and 1 explicitly in each search dimension.

        *Note - If input `times` is a float, then the start time of propagation is implicitly assumed to be zero.
        For an array of times, the first time point is the start time.*

        Returns
        -------
        OptimizationResult
            The result of the optimization.

        """
        if self._logger:
            self._logger.start()

        self._times = jnp.array([0.0, times]) if isinstance(times, float) else times

        self._build_optimizable_index_list()
        self._optimization_map.register_params_with_optimizables()

        # Collect the initial values of all parameters
        init = []
        for qty in self._optimization_map.get_all_parameters():
            init.append(qty.get_reduced_value())  # reduced values are between [-1, 1]

        opt_res = minimize(
            fun=self._set_parameters_and_measure,
            x0=np.concatenate(init).flatten(),
            bounds=[(-1, 1)] * self._opt_idxs[-1],
            method=self._method,
            options=self._options,
            callback=self._callback,
        )

        if self._logger:
            self._logger.stop(str(opt_res))

        return OptimizationResult(
            status=OptimizationResult.STATUS_SUCCESS if opt_res.success else OptimizationResult.STATUS_FAILED,
            value=opt_res.fun,
            iterations=opt_res.nfev,
            message=opt_res.message,
            raw_result=opt_res,
        )

    def _set_parameters_and_measure(self, values) -> float:
        """Update the parameter values and return the measurement result.

        Internal callback.

        Parameters
        ----------
        values: Array
            Parameter values for the update.

        Returns
        -------
        Array
            Returns the measurement result.

        """
        log = []
        params = self._optimization_map.get_all_parameters()
        for index, val in enumerate(np.split(values, self._opt_idxs[:-1])):
            params[index].set_reduced_value(val)
            log.append(params[index])
        infid = 1 - self._measure.calculate_normalized_scalar(self._times)

        if self._logger:
            self._logger.log(log, infid)
        return infid



"""Testing the GOAT optimization model."""

import numpy as np
import pytest

from paraqeet.measurement.state_transfer_fidelity import StateTransferFidelity
from paraqeet.measurement.unitary_fidelity import UnitaryFidelity
from paraqeet.model.closed_system import ClosedSystem
from paraqeet.model.drive_operator import DriveOperator
from paraqeet.model.open_system import OpenSystem
from paraqeet.model.qubit import Qubit
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.scipy_optimizer import ScipyOptimizer
from paraqeet.optimizers.scipy_optimizer_gradient import ScipyOptimizerGradient
from paraqeet.propagation.scipy_expm_goat import ScipyExpmGOAT
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import FlatTopGaussianEnvelope
from paraqeet.signal.iq_mixer import IQMixer

FREQ = 4.327884e9 * 2 * np.pi
T_FINAL = 13e-9

RES = 100e9
T1 = Quantity(10e-6, 1e-6, 100e-6)
TEMP = Quantity(10e-3, 1e-3, 50e-3)
T2STAR = Quantity(20e-6, 1e-6, 100e-6)


@pytest.fixture
def tone():
    """Return a cosine tone with a fixed error-function shaped envelope."""
    env = FlatTopGaussianEnvelope()
    env._t_up.set_value(T_FINAL / 5)
    env._t_down.set_value(4 * T_FINAL / 5)
    env._ramp_time.set_value(T_FINAL / 10)
    return env


@pytest.fixture
def gen(tone):
    """Generate a cosine tone."""
    gen = IQMixer(envelopes=[tone])
    return gen


@pytest.fixture(scope="function", params=["openSystem", "closedSystem"])
def prop(gen, request):
    """Solve the equation of motion.

    By piecewise exponentiation with the scipy package.

    """
    drive = DriveOperator(gen, is_longitudinal=False)
    controlled_qubit = Qubit(Quantity(FREQ, FREQ / 4, FREQ), drives=[drive], t1=T1, temp=TEMP, t2star=T2STAR)
    if request.param == "openSystem":
        model = OpenSystem(controlled_qubit)
    elif request.param == "closedSystem":
        model = ClosedSystem(controlled_qubit)
    return ScipyExpmGOAT(model=model, resolution=RES)


@pytest.fixture
def states(prop):
    """Compare the overlap of the initial and final state."""
    init = np.array([[1.0], [0.0]])
    target = np.array([[0.0], [1]])

    if prop.is_open:
        init = np.matmul(init, init.T)
        target = np.matmul(target, target.T)

    return StateTransferFidelity(
        propagation=prop,
        initial_state=init,
        target_state=target,
    )


@pytest.fixture
def gates(prop):
    """Compare the propagator with a gate via the L2 norm."""
    if prop.is_open:
        pytest.skip("Gate optimization is only implemented for closed system.")
    pauli_x = np.array([[0.0, 1], [1, 0.0]])
    prop.set_initial_state(np.identity(2))
    return UnitaryFidelity(
        propagation=prop,
        gate=pauli_x,
    )


@pytest.fixture
def opt_map(gen):
    """Create an optimization map."""
    params = gen.get_parameters()
    params[0].set_value(0.5 * np.pi / T_FINAL)
    params[-2].set_value(1.01 * FREQ)
    optmap = OptimizationMap()
    # Not optimizing t_up, t_down, ramp_time
    optmap.add(gen, [params[0], params[-2], params[-1]])
    return optmap


@pytest.fixture
def grad_opt(states, opt_map):
    """Create a scipy optimizer gradient object over states."""
    return ScipyOptimizerGradient(measure=states, optimization_map=opt_map)


@pytest.fixture
def grad_gates_opt(gates, opt_map):
    """Create a scipy optimizer gradient object over gates."""
    return ScipyOptimizerGradient(measure=gates, optimization_map=opt_map)


@pytest.fixture
def opt(states, opt_map):
    """Create a scipy optimizer object over states."""
    return ScipyOptimizer(measure=states, optimization_map=opt_map)


@pytest.fixture
def gates_opt(gates, opt_map):
    """Create a scipy optimizer gradient object over gates."""
    return ScipyOptimizer(measure=gates, optimization_map=opt_map)


def test_optim_finite_diff(opt) -> None:
    """Check that the optimization goes below threshold."""
    res = opt.optimize(times=T_FINAL)
    assert res.value < 1e-2


def test_optim_goat(grad_opt) -> None:
    """Check that the optimization goes below threshold."""
    res = grad_opt.optimize(times=T_FINAL)
    assert res.value < 1e-2


def test_optim_gates_finite_diff(gates_opt) -> None:
    """Check that the optimization goes below threshold."""
    res = gates_opt.optimize(times=T_FINAL)
    assert res.value < 1e-2


def test_optim_goat_gates(grad_gates_opt) -> None:
    """Check that the optimization goes below threshold."""
    res = grad_gates_opt.optimize(times=T_FINAL)
    assert res.value < 1e-2



"""Testing the device functions."""

from collections.abc import Callable
from functools import partial

import jax.numpy as jnp
from jax import jit
from jax.scipy.special import erf

from paraqeet.quantity import Array, Quantity
from paraqeet.signal.envelopes import Envelope


class FlatTopGaussianEnvelopeAD(Envelope):
    """A flat-top Gaussian envelope without analytic gradients.

    Dummy device to test AutoDiff Gradients for Envelopes.

    _amplitude: Quantity
        The amplitude of the envelope.
    _t_final: Quantity
        The length in time of the envelope.
    _gradient_function: Callable | None
        The function to calculate the gradient with respect to a set of
        previously defined parameters.
    _grad_arg_nums: tuple[int, ...]
        The identifying indices of which parameters to calculate the gradient
        with respect to.

    """

    _amplitude: Quantity
    _t_final: Quantity
    _t_up: Quantity
    _t_down: Quantity
    _ramp_time: Quantity

    def __init__(
        self,
        amplitude: Quantity | None = None,
        t_up: Quantity | None = None,
        t_down: Quantity | None = None,
        ramp_time: Quantity | None = None,
        t_final: Quantity | None = None,
    ):
        self._amplitude = amplitude or Quantity(
            1.55e8,
            min_value=jnp.array(0.0),
            max_value=jnp.array(1e9),
            unit="Hz",
            name="Amplitude",
            two_pi=True,
        )

        self._t_final = t_final or Quantity(
            32e-9,
            min_value=jnp.array(0),
            max_value=jnp.array(100e-9),
            unit="s",
            name="t_final",
        )

        self._t_up = t_up or Quantity(
            self._t_final.get_value() / 5,
            min_value=self._t_final.get_min_value(),
            max_value=self._t_final.get_max_value(),
            unit="s",
            name="t_up",
        )

        self._t_down = t_down or Quantity(
            4 * self._t_final.get_value() / 5,
            min_value=self._t_final.get_min_value(),
            max_value=self._t_final.get_max_value(),
            unit="s",
            name="t_down",
        )

        self._ramp_time = ramp_time or Quantity(
            self._t_final.get_value() / 10,
            min_value=self._t_final.get_min_value(),
            max_value=self._t_final.get_max_value(),
            unit="s",
            name="ramp_time",
        )

        self._gradient_function: Callable | None = None
        self._grad_arg_nums: tuple[int, ...] = ()

    def get_parameters(self):
        """Get all parameters of the system."""
        return [self._amplitude, self._t_up, self._t_down, self._ramp_time]

    @partial(jit, static_argnums=(0,))
    def _evaluate(self, amp: Array, t_up: Array, t_down: Array, ramp_time: Array, t: Array):
        """Compute the output of the device.

        Explicitly depends on the optimizable parameters.

        Parameters
        ----------
        amp: Quantity
            Cosine pulse amplitude.
        t_up: Quantity
            The start time of constant section of the envelope.
        t_down: Quantity
            The end time of constant section of the envelope.
        ramp_time: Quantity
            The rate of ramp up and ramp down of the envelope.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the output of the device that explicitly depends
            on the optimizable parameters.

        """
        ramp_up = 1 + erf((t - t_up) / ramp_time)
        ramp_down = 1 + erf((-t + t_down) / ramp_time)
        return amp * ramp_up * ramp_down / 4

    def get_value(self, times: Array | float) -> Array:
        """Get the output of the device on time stamps.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the output of the device.

        """
        amp = self._amplitude.get_value()
        t_up = self._t_up.get_value()
        t_down = self._t_down.get_value()
        ramp_time = self._ramp_time.get_value()
        # returns JitWrapped
        return self._evaluate(amp, t_up, t_down, ramp_time, times)  # type: ignore



"""Class definition of the Propagation model."""

from abc import ABC, abstractmethod

import jax.numpy as jnp
import numpy as np

from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.model.open_system import OpenSystem
from paraqeet.quantity import Array


class Propagation(ABC):
    """Abstract base class for any implementation of the equations of motion.

    The right-hand side of the equation is provided by the underlying model.

    Parameters
    ----------
    model: Model
        Represents the equation of motion for a given Hamiltonian.
    resolution: float
        Propagation resolution used to solve the equation of motion.
        The corresponding time step dt = 1/resolution
    """

    _model: EquationOfMotion | None
    _initial_state: Array | None = None
    _is_open: bool = False
    _resolution: float

    def __init__(self, model: EquationOfMotion | None, resolution: float):
        self._model = model
        self._resolution = resolution
        if isinstance(model, OpenSystem):
            self.is_open = True

    @property
    def is_open(self) -> bool:
        """Return if the propagation is for open or closed system."""
        return self._is_open

    @is_open.setter
    def is_open(self, flag) -> None:
        """Set if the propagation is for open or closed system."""
        self._is_open = flag

    @property
    def resolution(self) -> float:
        """Return the propagation resolution."""
        return self._resolution

    @resolution.setter
    def resolution(self, resolution: float) -> None:
        """Set the propagation resolution."""
        self._resolution = resolution

    def _construct_times(self, time, ti):
        """Construct one-dimensional vector of time.

        Interpolate the user-specified times to match the propagation resolution.

        Parameters
        ----------
        time: Array
            Array of timesteps.
        ti: int
            Snapshot of the time at a current step

        Returns
        -------
        Array
            Array of timestamps in specified resolution.
        int
            Difference in time step.

        """
        t0 = time[ti - 1]
        t1 = time[ti]
        steps = int(np.ceil((t1 - t0) * self.resolution))
        times = jnp.linspace(t0, t1, steps, endpoint=False)
        if steps < 2:
            dt = t1 - t0
        else:
            dt = times[1] - times[0]
        return times, dt

    # TODO: Keep or remove get_parameters?
    # @staticmethod
    def get_parameters(self):
        """Per default, propagation methods have no parameters to optimize."""
        return []

    def set_initial_state(self, state: Array):
        """Set the initial state for the propagation. (Default implementation)

        Propagation implementations that do not need the state should not
        implement this function.

        Subclasses can access the state in the _initial_sate field.

        Parameters
        ----------
        state: Array
            Parameter value to be set as the initial state for the propagation.

        """
        self._initial_state = state

    @abstractmethod
    def propagate(self, time: Array) -> Array:
        """Return the solution of the equations of motion.

        The first dimension of the result will always be the time.
        Like in the model, the format of the other dimensions depends on the
        implementation and could for example be a propagated state vector or
        a propagator in matrix form.

        Parameters
        ----------
        time: Array
            Any one-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the solution of the equations of motion.


        """
        # TODO: Distinguish between internal time (class property), i.e. the time grid of the
        # method vs. time points (input parameter) desired by other classes, e.g. Measurements
        pass



"""Testing the qubit optimizations."""

import numpy as np
import pytest

from paraqeet.logger import Logger
from paraqeet.measurement.state_transfer_fidelity import StateTransferFidelity
from paraqeet.model.closed_system import ClosedSystem
from paraqeet.model.drive_operator import DriveOperator
from paraqeet.model.qubit import Qubit
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.bayesian_optimizer import BayesianOptimizer
from paraqeet.optimizers.cmaes_optimizer import CMAEsOptimizer
from paraqeet.optimizers.scipy_optimizer import ScipyOptimizer
from paraqeet.propagation.scipy_expm_goat import ScipyExpmGOAT
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import ConstantEnvelope
from paraqeet.signal.iq_mixer import IQMixer

TONE = ConstantEnvelope()
GEN = IQMixer(envelopes=[TONE])
PARAMS = GEN.get_parameters()

FREQ = 4.8e9 * 2 * np.pi
T_FINAL = 10e-9

PARAMS[0].set_value(0.8 * np.pi / T_FINAL)
PARAMS[2].set_value(1.01 * FREQ)

DRIVE = DriveOperator(GEN, is_longitudinal=False)
CONTROLLED_QUBIT = Qubit(frequency=Quantity(FREQ, 0.8 * FREQ, 1.2 * FREQ), drives=[DRIVE])
MODEL = ClosedSystem(CONTROLLED_QUBIT)

PROP = ScipyExpmGOAT(MODEL, resolution=100e9)

INIT = np.array([[1.0], [0]])
TARGET = np.array([[0.0], [1]])
ZEROONE = StateTransferFidelity(
    propagation=PROP,
    initial_state=INIT,
    target_state=TARGET,
)


@pytest.fixture
def opt():
    """Create ScipyOptimizer optimizer."""
    optmap = OptimizationMap()
    optmap.add(GEN, [PARAMS[0], PARAMS[2]])
    return ScipyOptimizer(ZEROONE, optimization_map=optmap)


@pytest.fixture
def cma_opt():
    """Create CMAEs optimizer."""
    optmap = OptimizationMap()
    optmap.add(GEN, [PARAMS[0], PARAMS[2]])
    return CMAEsOptimizer(ZEROONE, optimization_map=optmap)


@pytest.fixture
def bay_opt():
    """Create Bayesian optimizer."""
    optmap = OptimizationMap()
    optmap.add(GEN, [PARAMS[0], PARAMS[2]])
    return BayesianOptimizer(ZEROONE, optimization_map=optmap)


def test_optim(opt) -> None:
    """Check that the optimization goes below threshold."""
    opt.logger = Logger()
    res = opt.optimize(times=T_FINAL)
    assert res.value < 1e-4


def test_cma(cma_opt: CMAEsOptimizer) -> None:
    """Check that the optimization goes below threshold."""
    cma_opt.logger = Logger()
    res = cma_opt.optimize(times=T_FINAL)
    assert res.value < 1e-4


def test_baysian(bay_opt: BayesianOptimizer) -> None:
    """Check that the optimization goes below threshold."""
    bay_opt.logger = Logger()
    bay_opt.iterations = 200
    res = bay_opt.optimize(times=T_FINAL)
    assert res.value < 1e-3



"""Test the state transfer fidelity model."""

import numpy as np
import pytest

from paraqeet.measurement.state_transfer_fidelity import (
    StateTransferFidelity,
    StateTransferFidelityAD,
)
from tests.propagation.identity_propagation import IdentityPropagation
from tests.propagation.random_propagation import RandomPropagation


@pytest.fixture
def identity_propagation():
    """Return a mock identity propagation object."""
    return IdentityPropagation()


def test_limits_vectors(random_state):
    """Test fidelity for state vectors is always in the interval [0, 1)."""
    for size in range(2, 30):
        initial_state = random_state(size)
        target_state = random_state(size)
        propagation = RandomPropagation(size, False)
        times = np.array([0.0, 1.0])
        measurement = StateTransferFidelity(
            propagation,
            initial_state,
            target_state,
        )

        for _ in range(20):
            m = measurement.measure(times=times)
            assert 0.0 <= m
            m = measurement.calculate_normalized_scalar(times=times)
            assert 0.0 <= m <= 1.0
            m, _ = measurement.get_value_and_gradient(times=times)
            assert 0.0 <= m


def test_vector_equality(identity_propagation, random_state):
    """Test that F(v,v) = 1 for state vectors."""
    for size in range(2, 30):
        for _ in range(100):
            state = random_state(size)
            identity_propagation.set_initial_state(state)
            measurement = StateTransferFidelity(identity_propagation, state, state)
            m = measurement.measure(times=np.array([1.0]))
            np.testing.assert_almost_equal(m, 1.0)


@pytest.mark.filterwarnings("ignore:Different shapes for")
def test_incompatible_shape(identity_propagation, random_state):
    """Test incompatible shapes for initial and target states.

    A set of initial and target states with different dimensions
    should raise an exception.

    Raises
    ------
    Exception
        Different dimensions for the initial and target states
        raise an exception.

    """
    allDims = np.arange(2, 30)
    for dim in allDims:
        for i in range(10):
            initialState = random_state(dim)
            dimensions = np.delete(allDims, np.where(allDims == dim)[0][0])
            targetState = random_state(np.random.choice(dimensions))

            fid = StateTransferFidelity(identity_propagation, initialState, targetState)

            fid_AD = StateTransferFidelityAD(identity_propagation, initialState, targetState)

            with pytest.raises(Exception):
                fid.measure(times=np.array([1.0]))
            with pytest.raises(Exception):
                fid_AD.measure(times=np.array([1.0]))



"""Class definition of Serializer."""

from abc import ABC, abstractmethod


class Serializer(ABC):
    """
    Interface for any class that can read and write configurations to a persistent format, e.g. a file. This can be used
    for the state of an optimization or the setup of the layers.
    """

    @abstractmethod
    def save(self, data: dict, comment: str | None = None) -> None:
        """
        Saves data to a persistent format. The actual format depends on the implementation.

        Parameters
        ----------
        data: dict
            the data to be exported
        comment: str
            Optional comment to be stored with the data, for example a description of the data. Implementations have
            to decide how to store the comment.
        -------

        """
        pass

    @abstractmethod
    def load(self) -> dict:
        """Loads and returns data that was previously saved."""
        pass

    @abstractmethod
    def load_comment(self) -> str | None:
        """Loads and returns the comment, if any, that was previously saved with the data. Returns None if no comment
        was saved.

        Returns
            The comment, or None if no comment was saved.
        """
        pass



"""Testing the Rabi Analytic Model."""

import jax
import pytest
from numpy.testing import assert_almost_equal

from paraqeet.measurement.rabi_experiment import RabiExperiment
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.scipy_optimizer import ScipyOptimizer

jax.config.update("jax_enable_x64", True)

FREQ = 4.8e9
RABI_NAME = "Analytic Rabi Model"
T_FINAL = 0.6e-9


@pytest.fixture
def opt(rabi):
    """Create optimization map with Rabi model."""
    optmap = OptimizationMap()
    optmap.add(rabi, rabi.get_parameters())
    return ScipyOptimizer(rabi, optmap)


@pytest.fixture
def rabi():
    """Create Rabi test object."""
    exp = RabiExperiment(FREQ)
    exp.name = RABI_NAME
    return exp


def test_name(rabi):
    """Check that the name is 'RABI_NAME'."""
    assert rabi.name == RABI_NAME


def test_rabi(opt) -> None:
    """Check that the rabi optimization goes below threshold."""
    res = opt.optimize(times=T_FINAL)
    assert res.value < 1e-7


def test_find_resonance(rabi, opt) -> None:
    """Check for resonance."""
    opt.optimize(times=T_FINAL)
    params = rabi.get_parameters()
    assert_almost_equal(FREQ / 1e9, params[1].get_value() / 1e9, decimal=4)



"""Class definition of the CMA-Es Optimizer model."""

from collections.abc import Callable

import cma.evolution_strategy as cma
import jax.numpy as jnp
import numpy as np

from paraqeet.file_logger import Logger
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.optimizer import OptimizationResult, Optimizer
from paraqeet.quantity import Array


class CMAEsOptimizer(Optimizer):
    """Wrapper for the pycma implementation of CMA-Es.

    The pycmi implementation has the following custom options for optimization:

    :param float noise: Artificial noise added to a function evaluation.

    :param boolean init_point: Force the use of the initial point in the first generation.

    :param float spread: Adjust the parameter spread of the first generation cloud.

    :param int stop_at_convergence: Custom stopping condition. Stop if the cloud shrunk for this number of generations.

    :param float stop_at_sigma: Custom stopping condition. Stop if the cloud shrunk to this standard deviation.

    See also: http://cma.gforge.inria.fr/apidocs-pycma/

    Parameters
    ----------
    measure: Measurement
        Represents any observable and the process of measurement itself.
    optimization_map: OptimizationMap
        Optimizable interface for all parameters considered in optimization.
    logger: FileLogger | None, default=None
        The file logger object.
    callback
        Callback function for optimization.

    """

    _measure: NormalizableMeasurement
    _options: dict
    _callback: Callable | None

    def __init__(
        self,
        measure: NormalizableMeasurement,
        optimization_map: OptimizationMap,
        logger: Logger | None = None,
        callback=None,
    ):
        super().__init__(measure, optimization_map, logger)
        self._options = {
            "noise": 0,
            "batch_noise": 0,
            "init_point": False,
            "spread": 0.1,
            "bounds": [-1.0, 1],
        }
        self.callback = callback

    @property
    def options(self) -> dict:
        """Get options from the system."""
        return self._options

    @options.setter
    def options(self, opts) -> None:
        """Set options for the system."""
        self._options.update(opts)

    @property
    def callback(self) -> Callable | None:
        """Returns the current callback function."""
        return self._callback

    @callback.setter
    def callback(self, cbfun: Callable) -> None:
        """Set the callback function for the optimizer.

        Parameters
        ----------
        Callable
            The function to be set as the callback.

        """
        self._callback = cbfun

    def optimize(self, times: Array | float) -> OptimizationResult:
        """Optimize the system via the CMA-Es optimizer.

        Performs the actual optimization via the following custom options:

        noise: float
            Artificial noise added to a function evaluation.
        init_point: boolean
            Force the use of the initial point in the first generation.
        spread: float
            Adjust the parameter spread of the first generation cloud.
        stop_at_convergence : int
            Custom stopping condition. Stop if the cloud shrunk for this number
            of generations.
        stop_at_sigma: float
            Custom stopping condition. Stop if the cloud shrunk to this
            standard deviation.

        *Note - If input `times` is a float, then the start time of propagation is implicitly assumed to be zero.
        For an array of times, the first time point is the start time.*

        Returns
        -------
        OptimizationResult
            Result of optimization via the OptimizationResult object.
            (status, value, iterations and the raw result)

        """
        self._times = jnp.array([0.0, times]) if isinstance(times, float) else times

        options = {}
        options.update(self._options)
        options = self._options
        if "noise" in options:
            noise = float(options.pop("noise"))

        if "batch_noise" in options:
            batch_noise = float(options.pop("batch_noise"))

        if "init_point" in options:
            init_point = bool(options.pop("init_point"))

        if "spread" in options:
            spread = float(options.pop("spread"))

        shrunk_check = False
        if "stop_at_convergence" in options:
            sigma_conv = int(options.pop("stop_at_convergence"))
            sigmas = []
            shrunk_check = True

        sigma_check = False
        if "stop_at_sigma" in options:
            stop_sigma = int(options.pop("stop_at_sigma"))
            sigma_check = True

        settings = options

        if self._logger:
            self._logger.start()

        self._build_optimizable_index_list()
        self._optimization_map.register_params_with_optimizables()

        x_init = []
        for qty in self._optimization_map.get_all_parameters():
            x_init.append(qty.get_reduced_value())

        es = cma.CMAEvolutionStrategy(np.concatenate(x_init).flatten(), spread, settings)
        iter = 0
        while not es.stop():
            if shrunk_check:
                sigmas.append(es.sigma)
                if iter > sigma_conv:
                    if all(sigmas[-(i + 1)] < sigmas[-(i + 2)] for i in range(sigma_conv - 1)):
                        print(f"ParaQeet:STATUS: Shrunk cloud for {sigma_conv} steps. Switching to gradients.")
                        break

            if sigma_check:
                if es.sigma < stop_sigma:
                    print("ParaQeet:STATUS: Goal sigma reached. Stopping CMA.")
                    break

            samples = es.ask()
            if init_point and iter == 0:
                samples.insert(0, x_init)
                print("ParaQeet:STATUS: Adding initial point to CMA sample.")
            solutions = []
            if batch_noise:
                error = np.random.randn() * noise
            for sample in samples:
                goal = self._set_parameters_and_measure(sample)
                if noise:
                    error = np.random.randn() * noise
                if batch_noise or noise:
                    goal = goal + error
                solutions.append(float(goal))
            es.tell(samples, solutions)
            es.disp()

            iter += 1
            if self._callback is not None:
                self._callback(samples)

        if self._logger:
            self._logger.stop(es.result_pretty())

        return OptimizationResult(
            status=self.__determine_termination_status(es.result.stop()),
            value=es.result.fbest,
            iterations=es.result.iterations,
            raw_result=es.result,
        )

    def _set_parameters_and_measure(self, values) -> float:
        """Update the parameter values and return the measurement result.

        Internal callback.

        Parameters
        ----------
        values: Array
            Values for the update of the parameters.

        Returns
        -------
        Array
            Returns the inverse of the fidelity.

        """
        log = []
        params = self._optimization_map.get_all_parameters()
        for index, val in enumerate(np.split(values, self._opt_idxs[:-1])):
            params[index].set_reduced_value(val)
            log.append(params[index])

        infid = 1.0 - self._measure.calculate_normalized_scalar(self._times)

        if self._logger:
            self._logger.log(log, infid)
        return infid

    def __determine_termination_status(self, conditions: dict) -> int:
        """Determine the optimization termination status.

        Determines the success or failure of the optimization depending on the
        termination conditions dict of the CMAEvolutionStrategy.

        Parameters
        ----------
        conditions: dict
            The dictionary from the CMAEvolutionStrategy.stop().

        Returns
        -------
        int
            One of the constants in OptimizationResult.

        """
        if any(
            (key in conditions)
            for key in [
                "ftarget",
                "tolfun",
                "tolfunhist",
                "tolfunrel",
                "tolfacupx",
                "tolx",
            ]
        ):
            return OptimizationResult.STATUS_SUCCESS
        elif any((key in conditions) for key in ["maxfevals", "maxiter", "timeout"]):
            return OptimizationResult.STATUS_FAILED
        else:
            # not decidable
            return OptimizationResult.STATUS_FINISHED



import numpy as np
import pytest

from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.model.resonator import Resonator
from paraqeet.model.rotating_frame_coupling import RotatingFrameCoupling


@pytest.fixture
def subsystem(random_quantity):
    @pytest.mark.usefixtures("random_quantity")
    def _method(dim: int) -> DifferentiableHamiltonian:
        frequency = random_quantity(1)
        return Resonator(dim, frequency)

    return _method


@pytest.fixture
def coupling(subsystem, random_quantity):
    @pytest.mark.usefixtures("random_quantity")
    def _method(dim1, dim2) -> RotatingFrameCoupling:
        sub1 = subsystem(dim1)
        sub2 = subsystem(dim2)
        coupling = random_quantity(1, "Hz")
        diff_frequency = random_quantity(1, "Hz")
        return RotatingFrameCoupling(sub1, sub2, coupling, diff_frequency)

    return _method


def test_get_parameters(coupling):
    coup = coupling(np.random.randint(2, 10), np.random.randint(2, 10))
    assert coup.get_parameters() is not None and len(coup.get_parameters()) >= 0


# Tests if get_matrices and get_matrices_one_time return matrices with the correct shape
def test_matrix_dimensions(coupling):
    dim1 = np.random.randint(2, 10)
    dim2 = np.random.randint(2, 10)
    dims = [dim1, dim2]
    coup = coupling(dim1, dim2)

    times = np.linspace(0.0, np.random.randint(1, 10) * np.random.rand(), np.random.randint(1, 10))
    mat = coup.get_RWA_couplings(times[-1])
    for i in range(len(mat)):  # iterate the coupling terms
        assert len(mat[i]) == len(dims)  # two subsystems
        for j in range(len(dims)):
            assert mat[i][j].shape == (dims[j], dims[j])


# Tests if get_gradient and get_gradient_one_time return matrices with the correct shape
def test_gradient_dimensions(coupling):
    dim1 = np.random.randint(2, 10)
    dim2 = np.random.randint(2, 10)
    dims = [dim1, dim2]
    coup = coupling(dim1, dim2)
    coup.set_optimizable_parameters(coup.get_parameters())

    times = np.linspace(0.0, np.random.randint(1, 10) * np.random.rand(), np.random.randint(1, 10))
    mat = coup.get_RWA_gradients_one_time(times[-1])
    for i in range(len(mat[0])):  # iterate the coupling terms
        assert len(mat[0][i]) == len(dims)  # two subsystems
        for j in range(len(dims)):
            assert mat[0][i][j].shape == (dims[j], dims[j])

    coup.set_optimizable_parameters([coup.get_parameters()[1]])
    mat = coup.get_RWA_gradients(times)
    for i in range(len(mat[0])):  # iterate the coupling terms
        assert len(mat[0][i]) == len(dims)  # two subsystems
        for j in range(len(dims)):
            assert mat[0][i][j].shape == (len(times), dims[j], dims[j])



"""Test the Makhlin Functional."""

import numpy as np
import pytest

from paraqeet.exceptions import ConfigurationException
from paraqeet.measurement.makhlin_functional import MakhlinFunctional
from tests.propagation.identity_propagation import IdentityPropagation
from tests.propagation.random_propagation import RandomPropagation

iswap = np.array([[1.0, 0, 0, 0], [0, 0, 1.0j, 0], [0, 1.0j, 0, 0], [0, 0, 0, 1.0]])
cnot = np.array([[1.0, 0, 0, 0], [0, 1.0, 0, 0], [0, 0, 0, 1.0], [0, 0, 1.0, 0]])
swap = np.array([[1.0, 0, 0, 0], [0, 0, 1.0, 0], [0, 1.0, 0, 0], [0, 0, 0, 1.0]])
sqrtSwap = np.array([[1.0, 0, 0, 0], [0, 1.0, 1.0, 0], [0, 1.0, 1.00, 0], [0, 0, 0, 1.0]])


def test_positivity():
    """Test the positivity.

    Whether distance is greater or equal to 0.

    """
    propagation = RandomPropagation(4, True)
    times = np.array([1.0])
    measurement = MakhlinFunctional(propagation)

    for i in range(100):
        m = measurement.measure(times)
        assert 0.0 <= m


def test_local_gates():
    """Test that local rotations have a distance of 2."""
    x = np.array(
        [
            [0, 1.0],
            [1.0, 0],
        ]
    )
    y = np.array(
        [
            [0, -1.0j],
            [1.0j, 0],
        ]
    )
    z = np.array([[1.0, 0], [0, -1.0]])
    gates1 = [np.kron(g, np.eye(2)) for g in [x, y, z]]
    gates2 = [np.kron(np.eye(2), g) for g in [x, y, z]]
    for gate in gates1 + gates2:
        propagation = IdentityPropagation()
        propagation.set_initial_state(gate)
        measurement = MakhlinFunctional(propagation)
        m = measurement.measure(np.array([1.0]))
        np.testing.assert_almost_equal(m, 2.0)


def test_perfect_entanglers():
    """Test that perfect entangling gates have a distance of 0."""
    for gate in [iswap, cnot]:
        propagation = IdentityPropagation()
        propagation.set_initial_state(gate)
        measurement = MakhlinFunctional(propagation)
        m = measurement.measure(np.array([1.0]))
        np.testing.assert_almost_equal(m, 0.0)


def test_invariants():
    """Test that special gates generate the expected Makhlin invariants."""
    expectedInvariants = [
        [np.eye(4), [1, 0, 3]],
        [iswap, [0, 0, -1]],
        [cnot, [0, 0, 1]],
        [swap, [-1, 0, -3]],
    ]

    propagation = IdentityPropagation()
    for gate, invariants in expectedInvariants:
        propagation.set_initial_state(gate)
        measurement = MakhlinFunctional(propagation, np.array(invariants))
        m = measurement.measure(np.array([1.0]))
        np.testing.assert_almost_equal(m, 0.0)


def test_incompatible_shape():
    """Test that all propagators which are not 4D raise an exception.

    Raises
    ------
    Exception
        Raise an error if all propagators are not 4D.

    """
    incompatibleDimensions = np.delete(np.arange(2, 30), 2)
    for dim in incompatibleDimensions:
        propagation = RandomPropagation(dim, True)
        measurement = MakhlinFunctional(propagation)
        with pytest.raises(Exception):
            _ = measurement.measure(np.array([1.0]))


def test_measurement_needs_time():
    propagation = RandomPropagation(4, True)
    measurement = MakhlinFunctional(propagation, None)
    with pytest.raises(ConfigurationException):
        measurement.measure(times=None)



"""Class definition of the State propagation model."""

from abc import abstractmethod

from paraqeet.differentiable import Differentiable
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array


# TODO: Does a differantiable propagation make sense physically? Should we rather have
# differantiable models and use composition instead?
class DifferentiablePropagation(Propagation, Differentiable):
    """Propagation methods that provide a get_value_and_gradient method.

    Parameters
    ----------
    model: Model
        Represents the equation of motion for a given Hamiltonian.
    """

    _initial_state: Array | None = None
    _is_open: bool = False
    _resolution: float

    def __init__(self, model: EquationOfMotion, resolution: float):
        super().__init__(model, resolution)

    @abstractmethod
    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Gradient method to be implemented

        Parameters
        ----------
        times: Array
            Array of timesteps.

        Returns
        -------
        tuple[Array, Array]
            First dimension is time, second dimension is the parameter.

        """
        pass



"""Class definition of the Weighted Sum Goal model."""

import itertools

import jax.numpy as jnp
import numpy as np

from paraqeet.differentiable import Differentiable
from paraqeet.exceptions import ConfigurationException
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.quantity import Array


class WeightedSumGoal(NormalizableMeasurement, Differentiable):
    """Combine multiple measurements into a single goal function.

    Parameters
    ----------
    measurements : list[Measurement]
        List of measurements.
    weights : Array
        List of weights.
    sum_of_squares_options : dict | None
        A dictionary that contains information about how to include the
        sum of square differences in the cost function. If not None
        then it must contain the following keys:

        - weight : float
            The weight of the sum of square differences
        - meas_bool : list[bool]
            A list of boolean of the same length as measurements. If
            an element is True then the corresponding measurement is included
            in the sum of square difference the goal function.
    measurement_in_sum_of_squares : list[Measurement] | None
        The list of measurements included in the sum of square difference cost
        function. It is None if sum_of_squares_options is None

    Raises
    ------
    ConfigurationException
        If number of measurements and weights are incompatible.
    UserWarning
        If the given weights are not normalized.

    """

    _measurements: list[NormalizableMeasurement]
    _weights: Array
    _sum_of_squares_options: dict | None
    _measurements_in_sum_of_squares: list[NormalizableMeasurement]

    def __init__(
        self, measurements: list[NormalizableMeasurement], weights: Array, sum_of_squares_options: dict | None = None
    ):
        self._measurements = measurements
        self._weights = weights
        self._sum_of_squares_options = sum_of_squares_options
        if len(measurements) != len(weights):
            raise ConfigurationException(
                f"Incompatible number of measurements {len(measurements)} and weights {{len(weights)}}"
            )
        if sum_of_squares_options is not None:
            expected_keys = ["weight", "meas_bool"]
            provided_keys = list(sum_of_squares_options.keys())

            for key in provided_keys:
                if key not in expected_keys:
                    raise KeyError("The dictionary must have weight and meas_bool as keys")

            if not np.isclose(sum(weights) + sum_of_squares_options["weight"], 1.0):
                raise UserWarning("Total supplied weights are not normalized.")

            self._measurements_in_sum_of_squares = [
                meas for meas, flag in zip(measurements, sum_of_squares_options["meas_bool"]) if flag
            ]
        else:
            if not np.isclose(sum(weights), 1.0):
                raise UserWarning("Supplied weights are not normalized.")
            self._measurements_in_sum_of_squares = []

    @property
    def measurements(self) -> list[NormalizableMeasurement]:
        """Returns the list of measurement"""
        return self._measurements

    @property
    def weights(self) -> Array:
        """Returns the weights used in the weighted goal function"""
        return self._weights

    @property
    def sum_of_square_options(self) -> dict | None:
        """Returns the dictionary with the options about the sum of square differences goal function"""
        return self._sum_of_squares_options

    @property
    def measurements_in_sum_of_squares(self) -> list[NormalizableMeasurement]:
        """Returns the list of measurement included in the sum of square difference cost function"""
        return self._measurements_in_sum_of_squares

    def measure(self, times: Array) -> Array | float:
        """Sum of plain weighted measurements.

        Returns
        -------
        Array
            Returns the plain weighted sum.

        """
        values = [m.measure(times=times) for m in self._measurements]
        sum_meas: Array | float = 0.0
        for ii, w in enumerate(self._weights):
            sum_meas += w * values[ii]
        if self._sum_of_squares_options is not None:
            sum_square_diff: Array | float = 0.0
            values_in_sum_of_squares = [
                val for val, flag in zip(values, self._sum_of_squares_options["meas_bool"]) if flag
            ]
            for meas_a, meas_b in itertools.combinations(values_in_sum_of_squares, 2):
                sum_square_diff += (meas_a - meas_b) ** 2
            sum_meas += self._sum_of_squares_options["weight"] * sum_square_diff
        return sum_meas

    def calculate_normalized_scalar(self, times: Array | float) -> float:
        """Sum of weighted measurements from normalized measurements.

        Returns
        -------
        Array
            Returns the normalized weighted sum.

        """
        values = [m.calculate_normalized_scalar(times=times) for m in self._measurements]
        sum_meas = 0.0
        for ii, w in enumerate(self._weights):
            sum_meas += w * values[ii]
        if self._sum_of_squares_options is not None:
            values_in_sum_of_squares = [
                val for val, flag in zip(values, self._sum_of_squares_options["meas_bool"]) if flag
            ]
            sum_square_diff = 0.0
            for meas_a, meas_b in itertools.combinations(values_in_sum_of_squares, 2):
                sum_square_diff += (meas_a - meas_b) ** 2
            sum_meas += self._sum_of_squares_options["weight"] * sum_square_diff
        return float(sum_meas)

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Sum of weighted measurements from gradient-ized measurements.

        Returns
        -------
        Array
            Returns the weighted sum wrt to gradients.
        Array
            Returns the sum of gradients.

        """
        # TODO: Move this to __init__, and store a differentiable_measurements: bool and check here.
        for mes in self._measurements:
            if not isinstance(mes, Differentiable):
                raise ConfigurationException(
                    "All measurements must be Differentiable to compute the gradient of the WeightedSumGoal"
                )

        # TODO: Remove the check by moving the check to init
        values_and_gradients = [
            m.get_value_and_gradient(times=times) for m in self._measurements if isinstance(m, Differentiable)
        ]
        sum_meas = jnp.array(0)
        sum_grads = jnp.zeros_like(values_and_gradients[0][1])
        for ii, w in enumerate(self._weights):
            sum_meas += w * values_and_gradients[ii][0]
            sum_grads += w * values_and_gradients[ii][1]
        if self._sum_of_squares_options is not None:
            values_and_gradients_in_sum_of_squares = [
                val_grad
                for val_grad, flag in zip(values_and_gradients, self._sum_of_squares_options["meas_bool"])
                if flag
            ]
            sum_square_diff = jnp.array(0)
            grads_diff = jnp.zeros_like(values_and_gradients_in_sum_of_squares[0][1])

            for meas_a, meas_b in itertools.combinations(values_and_gradients_in_sum_of_squares, 2):
                sum_square_diff += (meas_a[0] - meas_b[0]) ** 2
                grads_diff += 2 * (meas_a[0] - meas_b[0]) * (meas_a[1] - meas_b[1])
            sum_meas += self._sum_of_squares_options["weight"] * sum_square_diff
            sum_grads += self._sum_of_squares_options["weight"] * grads_diff
        return float(sum_meas), sum_grads



"""Class definition for a mixed state transfer fidelity model."""

import jax.numpy as jnp
import jax.scipy.linalg as sclin

from paraqeet.exceptions import IncompatibleLayersException
from paraqeet.measurement.measurement import Measurement
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array


class MixedStateTransferFidelity(Measurement):
    """Mixed state transfer fidelity measurement model.

    Fidelity measure that compares the overlap of the initial
    and final state of density matrices.
    Note: this implementation is still very inaccurate.

    Parameters
    ----------
    propagation : Propagation
        Abstract base class for any implementation that can solve
        the equation of motion.
    targetState : Array
        Final state of the density matrices.
    times : Array
        One-dimensional vector of timestamps.

    """

    _target_state: Array
    _target_state_sqrt: Array
    _propagation: Propagation

    def __init__(
        self,
        propagation: Propagation,
        targetState: Array,
    ):
        self._propagation = propagation
        self._target_state = targetState

        # store the sqrt of the density matrix to simplify the measurement
        self._target_state_sqrt = sclin.sqrtm(self._target_state)

    def measure(self, times: Array) -> Array | float:
        """Measure overlap between initial and final state of density matrices.

        Returns
        -------
        Array
            Overlap between initial and final state of density matrices.

        Raises
        ------
        IncompatibleLayersException
            Raises an exception if required vector shape is not received.

        """
        state = self._propagation.propagate(times)[-1]
        if state.shape != self._target_state.shape:
            raise IncompatibleLayersException(
                f"Need a state vector of size {self._target_state.shape}"
                "for the state transfer fidelity, "
                "but got shape {state.shape}"
            )

        # density matrix
        product = self._target_state_sqrt @ state @ self._target_state_sqrt
        return jnp.abs(jnp.trace(sclin.sqrtm(product))) ** 2



"""A plotting library for ParaQeet."""

import json

import jax.numpy as jnp
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib_inline.backend_inline
import numpy as np
from jax import vmap

from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array
from paraqeet.signal.generator import Generator
from paraqeet.signal.waveform import Waveform

matplotlib_inline.backend_inline.set_matplotlib_formats("png")

# Specifying the custom plotting fonts
mpl.rcParams["font.family"] = "serif"
mpl.rcParams["font.serif"] = "Tex Gyre Pagella"
mpl.rcParams["mathtext.fontset"] = "stix"
mpl.rcParams["font.size"] = " 10.0"
mpl.rcParams["axes.labelsize"] = " 11.0"
mpl.rcParams["legend.fontsize"] = "10.0"
mpl.rcParams["xtick.labelsize"] = "10.0"
mpl.rcParams["ytick.labelsize"] = "10.0"
mpl.rcParams["axes.linewidth"] = "1.0"
mpl.rcParams["xtick.major.size"] = "3.0"
mpl.rcParams["xtick.major.width"] = "1.0"
mpl.rcParams["ytick.major.size"] = "3.0"
mpl.rcParams["ytick.major.width"] = "1.0"
mpl.rcParams["axes.titlesize"] = "medium"
mpl.rcParams["figure.titlesize"] = "medium"
plt.rcParams["figure.dpi"] = 125  # Set default DPI for all figures
plt.rcParams["savefig.dpi"] = 300  # Higher DPI for saved files

# Specifying a custom color palette
default_colors = mpl.rcParams["axes.prop_cycle"].by_key()["color"]
custom_colors = [
    "#18428A",
    "#FF9258",
    "#57B870",
    "#BF433B",
    "#FF00A6",
    "#FFCC33",
    "#008B8B",
    "#9B5DE5",
] + default_colors

mpl.rcParams["axes.prop_cycle"] = mpl.cycler(color=custom_colors)  # type: ignore


def plot_signal_and_dynamics(
    generator: Generator,
    propagation: Propagation,
    times: Array,
    state_labels: list[str] | None = None,
    axes=None,
    linestyle: str = "-",
    label: str = "",
    alpha: float = 1.0,
    linewidth: float = 1.5,
    marker: str = "",
):
    """Plot the signal and the corresponding dynamics.

    If fig or ax is provided then ax[0] is used to plot the signal and ax[1] for dynamics.
    This can be used to plot multiple signals and dynamics on the same plot.
    """

    def calculate_populations(states, dm=False):
        """Calculate state populations from density matrices and vectorized dm."""
        if len(states.shape) > 2:
            if dm:
                pops = jnp.abs(vmap(jnp.diag, in_axes=0)(states))
            else:
                pops = jnp.abs(states[:, :, 0]) ** 2
                pops = jnp.reshape(pops, [pops.shape[0], pops.shape[1]])
        else:
            if dm:
                pops = jnp.diag(states)
            else:
                pops = jnp.abs(states) ** 2
        return pops

    states = propagation.propagate(times)
    sig = generator.get_value(times) / 1e6 / (2 * np.pi)

    if axes is None:
        _, axes = plt.subplots(2, figsize=(4, 5), sharex=True)

    axes[0].plot(
        times / 1e-9, np.real(sig), label="Re " + label, ls=linestyle, alpha=alpha, linewidth=linewidth, marker=marker
    )
    axes[0].plot(
        times / 1e-9, np.imag(sig), label="Imag " + label, ls=linestyle, alpha=alpha, linewidth=linewidth, marker=marker
    )
    axes[0].legend(loc=1)
    axes[0].set_ylabel("Amplitude \n" + r"[MHz / $2\pi$]")
    axes[0].grid(True, linestyle=(1, (1, 5)), linewidth=1)

    axes[1].plot(
        times / 1e-9,
        calculate_populations(states, dm=propagation._is_open),
        ls=linestyle,
        alpha=alpha,
        linewidth=linewidth,
        marker=marker,
    )
    axes[1].set_ylabel("Population")
    axes[-1].set_xlabel("Time [ns]")
    if state_labels is not None:
        axes[1].legend(state_labels)
    axes[1].grid(True, linestyle=(1, (1, 5)), linewidth=1)

    return axes


def plot_signal(
    device: Generator | Waveform,
    times: Array,
    axes=None,
    linestyle: str = "-",
    label: str = "",
    alpha: float = 1.0,
    linewidth: float = 1.5,
    marker: str = "",
):
    """Plot signal from Generator or Envelope."""
    if isinstance(device, Generator):
        sig = device.get_value(times) / 1e6 / (2 * np.pi)
    else:
        sig = device.get_value(times) / 1e6 / (2 * np.pi)

    if axes is None:
        _, axes = plt.subplots(1, figsize=(4, 3))

    axes.plot(
        times / 1e-9,
        np.real(sig),
        label="Re " + label,
        ls=linestyle,
        alpha=alpha,
        linewidth=linewidth,
        marker=marker,
    )
    axes.plot(
        times / 1e-9,
        np.imag(sig),
        label="Imag " + label,
        ls=linestyle,
        alpha=alpha,
        linewidth=linewidth,
        marker=marker,
    )
    axes.grid(True, linestyle=(1, (1, 5)), linewidth=1)
    axes.set_ylabel(r"Amplitude [MHz / $2\pi$]")
    axes.set_xlabel("Time [ns]")
    axes.legend()
    return axes


def plot_infidelity_vs_evaluation_from_logs(
    log_path: str,
    axes=None,
    linestyle: str = "-",
    label: str = "",
    alpha: float = 1.0,
    linewidth: float = 1.5,
    marker: str = "",
):
    """Plot Infidelity vs Evaluation from json log file.

    Parameters
    ----------
    log_path : str
        Path to json log file.
    """
    with open(log_path) as file:
        log = [json.loads(line) for line in file]

    infidelities = []
    for eval in log:
        infidelities.append(eval["Goal"])

    if axes is None:
        _, axes = plt.subplots(1, figsize=(4, 4), sharex=True)

    axes.plot(
        range(len(infidelities)),
        infidelities,
        label=label,
        ls=linestyle,
        alpha=alpha,
        linewidth=linewidth,
        marker=marker,
    )
    axes.axhline(np.min(infidelities), color="red", ls="--", alpha=0.7)
    axes.set_yscale("log")
    axes.set_ylim(bottom=0.7 * min(infidelities))
    axes.text(x=len(infidelities) * 0.1, y=0.8 * min(infidelities), s=f"Minimum infidelity = {min(infidelities):.3e}")

    axes.legend(loc=1)
    axes.set_ylabel("Infidelity")
    axes.set_xlabel("Evaluation number")
    axes.grid(True, linestyle=(1, (1, 5)), linewidth=1)
    return axes



"""Custom Exception implementations."""


class IncompatibleLayersException(Exception):
    """Raise when incompatible layers transmission.

    Raised when a layer can not handle the form of the result of the
    previous layer. For example, a unitary fidelity will throw this if the
    propagation layer only provides the propagated state.

    """

    pass


class ConfigurationException(Exception):
    """Raise when configuration issue.

    Raised when a layer implementation was not properly configured before
    running it.

    """

    pass


class IncompatibleQuantityException(Exception):
    """Raise when incompatible quantity shape.

    Raised when a quantity has an unexpected shape, e.g. a vector quantity
    if a scalar was expected.

    """

    pass


class IncompatibleOptimizationMap(Exception):
    """Raise when incorrect number of quantities are specified.

    Raised when the number of quantities specified in optimization map
    doesn't match the number of gradients computed.

    """

    pass


class SerializationException(Exception):
    """Raised when reading or writing of a Quantity or an OptimizationMap fails."""

    pass



"""Class definition of a qubit model."""

import jax.numpy as jnp

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.model.drive import Drive
from paraqeet.quantity import Array, Quantity


# TODO: Why does Qubit inherit from Hamiltonian? Is a qubit really an operator? 
# What does a derivative of a qubit mean physically? Does get_value_at_timestep
# mean a measurement of the qubit at a certain time? 
class Qubit(DifferentiableHamiltonian):
    """Hamiltonian of a single qubit frequency/2 * sigma_z.

    The implementation uses the convention of having the excited state
    of the qubit as the first entry in the state. If you need a two-level
    system that is compatible with the projection of a higher-dimensional
    system (ground state as first entry), use a resonator and restrict its
    dimension to 2.

    Parameters
    ----------
    frequency : Quantity
        Frequency for characterizing the qubit.
    drives : list[Drive] | None
        List of time-dependent drives.

    """

    _frequency: Quantity
    _annihilation_op: Array
    _drift: Array
    _t1: Quantity | None
    _temp: Quantity | None
    _t2star: Quantity | None

    # TODO: should we move DifferentiableHamiltonian-object to the constructor instead of inheriting from it?
    def __init__(
        self,
        frequency: Quantity,
        drives: list[Drive] | None = None,
        t1: Quantity | None = None,
        temp: Quantity | None = None,
        t2star: Quantity | None = None,
    ):
        super().__init__(drives)
        self._frequency = frequency
        self._annihilation_op = jnp.array(
            [
                [0.0, 0.0],
                [1.0, 0.0],
            ]
        )
        self._drift = 0.5 * jnp.diag(jnp.array([1.0, -1.0]))
        self.t1 = t1
        self.temp = temp
        self.t2star = t2star

    @property
    def frequency(self) -> Quantity:
        """Get the frequency of the qubit."""
        return self._frequency

    @frequency.setter
    def frequency(self, frequency: Quantity) -> None:
        """Set the frequency of the qubit."""
        self._frequency = frequency

    @property
    def t1(self) -> Quantity | None:
        """Get the t1 of the resonator."""
        return self._t1

    @t1.setter
    def t1(self, t1: Quantity | None) -> None:
        """Set the t1 of the resonator."""
        self._t1 = t1

    @property
    def temp(self) -> Quantity | None:
        """Get the temp of the resonator."""
        return self._temp

    @temp.setter
    def temp(self, temp: Quantity | None) -> None:
        """Set the temp of the resonator."""
        self._temp = temp

    @property
    def t2star(self) -> Quantity | None:
        """Get the t2star of the resonator."""
        return self._t2star

    @t2star.setter
    def t2star(self, t2star: Quantity | None) -> None:
        """Set the t2star of the resonator."""
        self._t2star = t2star

    def get_parameters(self) -> list[Quantity]:
        """Get parameters of the model.

        Returns
        -------
        list[Quantity]
            Returns the list of parameters of the system.

        """
        return self._get_drive_parameters() + [self._frequency]

    def dimension(self) -> int:
        """Dimension of the qubit.

        Returns
        -------
        int
            Returns 2 as the dimension.

        """
        return 2

    def get_value_at_timestep(self, timestep: float) -> Array:
        """Get the drive matrix.

        Parameters
        ----------
        timestep: Array
            One time stamp.

        Returns
        -------
        Array
            The repeated drive matrix.

        """
        hamil = self._frequency.get_value() * self._drift
        return hamil + self._get_drive_matrix_at_timestep(self._annihilation_op, timestep)

    def get_gradient_at_timestep(self, time: float) -> Array:
        """Get the matrix representations of value and gradient of the drive as a tuple.

        Parameters
        ----------
        times: Array
            Array of timestamps of interest.

        Returns
        -------
        tuple[Array, Array]
            Returns the value und gradients of the drive.

        """
        # Fetch the gradient of the drive
        derivatives = self._get_drive_gradients_at_timestep(self._annihilation_op, time)

        # Combine with the derivative wrt the frequency
        if self._is_optimized(self._frequency):
            hamil = self._drift.reshape((1, 2, 2))
            derivatives = jnp.append(derivatives, hamil, axis=0)
        return derivatives

    def get_decay_rates(self) -> list[Array]:
        """Return decay rate for T1, T2star and Temp respectively."""
        if (self.t1 is None) or (self.t2star is None) or (self.temp is None):
            raise ConfigurationException("Specify values of T1, T2star and Temp for Open system simulations.")

        gamma = 1 / self.t1.get_value()
        gamma_t2star = 0.5 / self.t2star.get_value()

        hbar_over_kb = 7.638232582257738e-12
        beta = hbar_over_kb / (self.temp.get_value())
        nbar = jnp.exp(-beta * self.frequency.get_value())
        gamma_temp = gamma * nbar
        gamma_t1 = gamma * (nbar + 1)
        return [gamma_t1, gamma_temp, gamma_t2star]

    def get_collapseops(self) -> list[tuple[Array, Array]]:
        """Return a list tuples of decay rates and collapse operators for each subsystem."""
        gamma_t1, gamma_temp, gamma_t2star = self.get_decay_rates()
        col_t1 = self._annihilation_op
        col_temp = self._annihilation_op.T
        col_t2star = 2 * jnp.matmul(self._annihilation_op.T, self._annihilation_op)
        return [(gamma_t1, col_t1), (gamma_temp, col_temp), (gamma_t2star, col_t2star)]



"""Class definition of a Rabi experiment model."""

import jax.numpy as jnp

from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Array, Quantity


class RabiExperiment(NormalizableMeasurement, Optimizable):
    """Analytic model of the general Rabi formula.

    Parameters
    ----------
    qubit_freq : Quantity
        Resonance of the single qubit.

    """

    _qubit_freq: Quantity
    _amp: Quantity
    _freq: Quantity

    def __init__(self, qubit_freq: float) -> None:
        self._qubit_freq = Quantity(qubit_freq, 0.0, 10e9)
        self._amp = Quantity(60e6, 0, 100e6, "Hz")
        self._freq = Quantity(0.6 * qubit_freq, 0, 10e9)

    def get_parameters(self) -> list[Quantity]:
        """Return a list of parameters accessible in this measurement.

        Returns
        -------
        List[Quantity]
            List of parameters accessible in this measurement.

        """
        return [self._amp, self._freq]

    def measure(self, times: Array) -> Array | float:
        """Return measurement in the range [0, 1]."""
        return self.calculate_normalized_scalar(times)

    def calculate_normalized_scalar(self, times: Array | float) -> float:
        """Carry out a measurement operation.

        Gives the result of a general Rabi oscillation,
        depending of drive frequency, amplitude and time.

        *Note: Returns the measumement value at the last time point.*

        Returns
        -------
        Array
            Result of a general Rabi oscillation.

        """
        t = times if isinstance(times, float) else times[-1]
        q_freq = self._qubit_freq.get_value()
        amp = self._amp.get_value() * 2 * jnp.pi
        freq = self._freq.get_value()
        diff_sq = (q_freq - freq) ** 2
        norm = jnp.sqrt(1 + diff_sq / (amp**2))
        return float(jnp.abs(jnp.cos(jnp.sqrt(diff_sq + amp**2) / 2 * t) / norm**2).item())



"""Class definition of the Drive Hamiltonian in the rotating frame of drive."""

import jax.numpy as jnp

from paraqeet.model.drive import Drive
from paraqeet.quantity import Array, Quantity
from paraqeet.signal.generator import Generator


class RotatingFrameDrive(Drive):
    """Drive Hamiltonian in the Frame rotating at the frequency of the drive.

    _signal_generator: Generator
        Signal Generator without a LO, like the PWCGenerator
    """

    _signal_generator: Generator

    def __init__(self, signal_generator: Generator):
        self._signal_generator = signal_generator

    @property
    def generator(self) -> Generator:
        """Get the signal generator from the system.

        Returns
        -------
        Generator
            Returns the signal generator object from the system.

        """
        return self._signal_generator

    def get_parameters(self) -> list[Quantity]:
        """Get a list of parameters of the system.

        Returns
        -------
        list[Quantity]
            List of optimizable parameters of the system.

        """
        return self.generator.get_parameters()

    def get_value_at_timestep(self, annihilation_operator: Array, t: float) -> Array:
        r"""Implement drive in the rotating frame of drive.

        Drive Hamiltonian is implemented as
        \\big\\{ \\Omega a + \\Omega^* a^\\dagger \\big\\}
        Where \\Omega is the envelope (without the LO).

        Parameters
        ----------
        annihilation_operator: Array
            Annihilation operator of the subsystem
        t: Array
            One time step
        """
        env = self.generator.get_value(jnp.array([t]))
        return env * annihilation_operator + jnp.conjugate(env) * annihilation_operator.conj().T

    def get_gradient_at_timestep(self, annihilation_operator: Array, timestep: float) -> Array:
        """Get the one-time gradient of the system.

        Fetches the gradient from the drive and transforms it into the
        correct shape for the Hamiltonian.

        Parameters
        ----------
        annihilation_operator: Array
            Operator for longitudinal or transverse drive.
        timestep: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the shape-shifted gradient from the drive.

        """
        _, env_grad = self.generator.get_value_and_gradient(jnp.array([timestep]))
        env_grad = env_grad.reshape((-1, 1, 1))
        return env_grad * annihilation_operator + jnp.conjugate(env_grad) * annihilation_operator.conj().T



"""Class definition of the file logger object."""

import json
import os

from paraqeet.logger import Logger
from paraqeet.quantity import Quantity


class FileLogger(Logger):
    """Logger that writes messages to a file.

    Parameters
    ----------
    logdir: str="."
        Destination directory to store the logs.

    """

    _logdir: str
    _logfile: str
    _result_file: str

    def __init__(self, logdir: str = ".") -> None:
        self.logdir = logdir

    @property
    def logdir(self) -> str:
        """Returns the current log directory."""
        return self._logdir

    @logdir.setter
    def logdir(self, logdir: str):
        """Set the destination log directory.

        Stores both the log and the result files.

        Parameters
        ----------
        logdir: str
            Destination directory to store the logs.

        """
        self._logdir = logdir
        self._logfile = os.path.join(self._logdir, "opt.log")
        self._result_file = os.path.join(self._logdir, "opt.result")
        if not os.path.isdir(self._logdir):
            os.makedirs(self._logdir)

    def start(self):
        """Start logging."""
        super().start()

    def log(self, params: list[Quantity], infidelity: float):
        """Write the formatted parameters and the goal to the log file.

        Parameters
        ----------
        params: list[Quantity]
            List of quantities to be written to the log file.
        infidelity: float
            Goal value to be written to the log file.

        """
        super().log(params, infidelity)
        formatted_params = [param.get_value().tolist() for param in params]
        status = {
            "Eval": self._counter,
            "Parameters": formatted_params,
            "Goal": infidelity,
        }
        with open(self._logfile, "a") as log:
            log.write(json.dumps(status))
            log.write("\n")
            log.flush()

    def stop(self, result_message: str | None = None):
        """Stop logging and end the log file with the run information.

        Parameters
        ----------
        result_message: str | None = None
            The message that the user wants to write at the end of the log file.

        """
        super().stop()
        with open(self._result_file, "a") as log:
            if result_message:
                log.write(result_message)
                log.write("\n")
            log.write(f"Finished at {self._stop_time}\n")
            log.write(f"Total runtime: {self._stop_time - self._start_time}")
            log.write("\n")
            log.flush()



"""Class definition of the Makhlin functional."""

import jax.numpy as jnp

from paraqeet.exceptions import (
    ConfigurationException,
    IncompatibleLayersException,
)
from paraqeet.measurement.measurement import Measurement
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array


class MakhlinFunctional(Measurement):
    """Class definition of the Makhlin Functional invariants.

    Measures the distance of a propagator to a perfect entangler
    using Makhlin invariants.
    If a list of ideal Makhlin invariants is given,
    the distance is measured as the Euclidean distance between
    the actual and ideal invariants.
    Else, the Makhlin distance is used.

    Parameters
    ----------
    propagation : Propagation
        Abstract base class for any implementation that can solve
        the equation of motion.
    times : Array
        One-dimensional vector of timestamps.
    ideal_invariants : Array optional
        One-dimensional vector of ideal Makhlin invariants.

    """

    _propagation: Propagation
    _ideal_invariants: Array | None

    def __init__(
        self,
        propagation: Propagation,
        ideal_invariants: Array | None = None,
    ):
        self._propagation = propagation
        self._ideal_invariants = ideal_invariants

    def measure(self, times: Array) -> Array | float:
        """Measure distance of the propagator to a perfect entangler.

        Parameters
        ----------
        times : Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Distance of propagator.

        Raises
        ------
        IncompatibleLayersException
            Raises an exception if a quadratic unitary
            4x4 operator is not received.

        """
        if not times:
            raise ConfigurationException("Time array was not specified")

        u = self._propagation.propagate(times)[-1]
        if u.shape != (4, 4):
            raise IncompatibleLayersException("quadratic unitary 4x4 propagator needed for Makhlin invariants")
        gs = self._makhlin_invariants(u)
        if self._ideal_invariants is not None:
            return jnp.array(jnp.linalg.norm(gs - self._ideal_invariants))
        else:
            return jnp.abs(gs[2] * jnp.sqrt(gs[0] ** 2 + gs[1] ** 2) - gs[0])

    def _makhlin_invariants(self, u: Array) -> tuple[Array, Array, Array]:
        """Compute the Makhlin invariants for a matrix U.

        Returns a tuple with the three invariants g1, g2 and g3.

        Parameters
        ----------
        U: Array
            Input matrix for computing the Makhlin invariants of.

        Returns
        -------
        Tuple[Array, Array, Array]
            Returns a tuple of 3 Numpy Array as invariants g1, g2 and g3.

        """
        # transform to bell basis
        q = jnp.array(
            [[1, 0, 0, 1j], [0, 1j, 1, 0], [0, 1j, -1, 0], [1, 0, 0, -1j]],
        )
        det = jnp.linalg.det(u)
        # Normalize the determinant to be sensitive to leakage, non-unitarity.
        if det != 0.0:
            det /= jnp.abs(det)
        u_b = (q.T.conj() @ u @ q) / 2
        m = u_b.T @ u_b
        tr = jnp.trace(m @ m)
        tr_sq = jnp.trace(m) ** 2 / det
        return (
            jnp.real(tr_sq) / 16,
            jnp.imag(tr_sq) / 16,
            jnp.real(tr_sq - tr / det) / 4,
        )



"""Class definition of the Generator Drive model."""

import jax.numpy as jnp

from paraqeet.model.drive import Drive
from paraqeet.quantity import Array, Quantity
from paraqeet.signal.generator import Generator


class DriveOperator(Drive):
    r"""Create a generator drive model.

    Transversal (a^\\dagger + a or \\sigma_x type) or longitudinal (a^\\dagger a or \\sigma_z type) drive with a
    time-dependent scalar coefficient that is generated by a Generator object.

    Parameters
    ----------
    generator : Generator
        Signal generator stack.
    is_longitudinal : bool
        Generator is longitudinal or transversal depending on this boolean.

    """

    _generator: Generator
    _is_longitudinal: bool

    def __init__(self, generator: Generator, is_longitudinal: bool):
        self._generator = generator
        self._is_longitudinal = is_longitudinal

    @property
    def generator(self) -> Generator:
        """Get the signal generator from the system.

        Returns
        -------
        paraqeet.signal.generator.Generator
            Returns the signal generator object from the system.

        """
        return self._generator

    def get_parameters(self) -> list[Quantity]:
        """Get a list of parameters of the system.

        Returns
        -------
        list[Quantity]
            List of optimizable parameters of the system.

        """
        return self._generator.get_parameters()

    def _compute_matrix(self, a: Array) -> Array:
        """Return the operator for the longitudinal or transverse drive.

        Parameters
        ----------
        a: Array
            Operator for computation.

        Returns
        -------
        Array
            Returns the operator for the longitudinal or transverse drive.

        """
        return (jnp.conjugate(a.T) @ a) if self._is_longitudinal else (jnp.conjugate(a.T) + a)

    def get_value_at_timestep(self, a: Array, t: float) -> Array:
        """Get the one-time matrix of the system.

        Fetches the coefficient from the drive and transforms it
        into the correct shape for the Hamiltonian.

        Parameters
        ----------
        a: Array
            Operator for longitudinal or transverse drive.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the shape-shifted coefficient from the drive.

        """
        # TODO: generator.get_value expects an array, even for one time point.
        # Is the naming of the method correct then?
        signal = self._generator.get_value(t)
        matrix = self._compute_matrix(a)
        return signal * matrix

    def get_gradient_at_timestep(self, a: Array, timestep: float) -> Array:
        """Get the one-time gradient of the system.

        Fetches the gradient from the drive and transforms it into the
        correct shape for the Hamiltonian.

        Parameters
        ----------
        a: Array
            Operator for longitudinal or transverse drive.
        timestep: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the shape-shifted gradient from the drive.

        """
        signal_grad = self._generator.get_gradient_at_timestep(timestep).reshape((-1, 1, 1))
        matrix = Drive._repeat(self._compute_matrix(a), signal_grad.shape[0])
        return signal_grad * matrix



"""Tests for the utility functions in the abstract Measurement class.

These tests should be independent of specific implementations.
"""

import numpy as np
import pytest

from tests.measurement.random_measurement import RandomMeasurement
from tests.propagation.random_propagation import RandomPropagation


def random_state(dimension):
    """Generate random state according to the given dimension value."""
    state = np.random.random((dimension, 1)) + 1j * np.random.random((dimension, 1))
    return state / np.sqrt(np.vdot(state, state))


@pytest.mark.filterwarnings("ignore:Different shapes for")
def test_limit_projected_vectors(random_state):
    """Test the projection to a subspace.

    The projection to a subspace should not increase the
    range of possible measurement outcomes for state vectors.

    """
    times = np.array([1.0])
    for size in range(3, 30):
        for projectedSize in range(2, size):
            propagation = RandomPropagation(size, False)
            measurement = RandomMeasurement(propagation=propagation)
            for _ in range(20):
                m = measurement.measure(times=times)
                assert 0.0 <= m <= 1.0


def test_gate_shape(random_unitary_matrix):
    """Test the gate shape.

    The projection to a subspace should not increase the
    range of possible measurement outcomes for propagators.

    """
    times = np.array([1.0])
    for size in range(5, 30):
        for projectedSize in range(2, size):
            propagation = RandomPropagation(size, True)
            measurement = RandomMeasurement(propagation=propagation)
            for _ in range(20):
                m = measurement.measure(times=times)
                assert 0.0 <= m <= 1.0



"""Class definition of the Measurement model."""

from abc import ABC, abstractmethod

from paraqeet.quantity import Array


class Measurement(ABC):
    """Represents any observable and the process of measurement itself.

    The observable is measured after the propagation class
    has solved the equation of motion.

    Parameters
    ----------
    times: Array | None, optional
        One-dimensional vector of timestamps.

    """

    @abstractmethod
    def measure(self, times: Array) -> Array | float:
        """Measure the observable and returns the value.

        Parameters
        ----------
        times : Array
            One-dimensional vector of timestamps.
        projector : Array | None
            The projector matrix to restrict the operator.

        Returns
        -------
        Array or float
            This abstract method must return an Array or a float when
            implemented by subclasses. Might return multiple values.


        """
        pass


class NormalizableMeasurement(Measurement):
    """An abstract class for measurements providing normalized scalar value.

    Subclasses must implement the calculate_normalized_scalar() method which would
    return a measured value between 0 and 1.
    """

    @abstractmethod
    def calculate_normalized_scalar(self, times: Array | float) -> float:
        """Measure the normalized observable.

        Returns a single scalar value between 0 and 1.
        This function must be implemented by subclasses.

        Parameters
        ----------
        times : Array
            One-dimensional vector of timestamps.
        projection : Array | None
            The projector matrix to restrict the operator.

        Returns
        -------
        float
            Returns a float if implemented by a subclass.

        """
        pass



"""Class definition for the Quantity model."""

from __future__ import annotations  # necessary for type hints

import copy
from collections.abc import Callable
from sys import float_info
from typing import Self

import jax
import jax.numpy as jnp
import numpy as np

from paraqeet.exceptions import IncompatibleQuantityException

type Array = np.typing.NDArray[np.float64] | np.typing.NDArray[np.complexfloating] | jax.Array
jax.config.update("jax_enable_x64", True)


class Quantity:
    r"""Represent any physical quantity used in model or pulse specification.

    For arithmetic operations just the numeric value is used.
    The value itself is stored in an optimizer friendly way as a float
    between -1 and 1. The conversion is given by
    scale * (value + 1) / 2 + offset

    For convenience, the constructor and setter functions accept primitive floats. However, these will be converted into
    numpy arrays internally, such that scalar values are represented by arrays of shape (1,). All getter functions only
    return numpy arrays. If the value is an array and min/max are floats, the latter will be considered constant bounds
    for all value and will be converted into constant arrays.

    Note on python's operators: equality checks `q == p` and `q != p` check
    for the values of the quantities q and p. For vector or matrix quantities,
    these check if all values are equal. If you want to be sure that two
    quantities are the same object (i.e. the same memory address), use `q is p`.
    Ordering operators like `q > p` will only work for scalar quantities and
    will raise an exception for vector or matrix quantities.

    Parameters
    ----------
    value: Array | float
        Value of the quantity
    min_value: Array | float
        Minimum this quantity is allowed to take.
        If this is a float, it will be a default interval around the value will be chosen.
    max_value: Array | float
        Maximum this quantity is allowed to take.
    unit: str
        physical unit
    name: str
        symbol or description of this quantity
    two_pi: bool
        divide by two pi for representation

    Raises
    ------
    IncompatibleQuantityException
        If misconfigured by the user, e.g., bounds are not given or the wrong shape.

    """

    _unit: str
    _name: str
    _length: int
    _shape: tuple[int, ...]
    # internal representation of the value
    _value: Array
    _offset: Array
    _scale: Array
    _two_pi: bool
    _dependent: bool
    _dependencies: list
    _relation: Callable | None
    _dependents: list

    def __init__(
        self,
        value: Array | float,
        min_value: Array | float,
        max_value: Array | float,
        unit: str = "",
        name: str = "",
        two_pi: bool = False,
    ):
        if value is None or max_value is None or min_value is None:
            raise IncompatibleQuantityException("value, minimum, and maximum must be not null")

        self._unit = unit
        self._name = name
        self._scale = jnp.array(0)
        self._two_pi = two_pi

        value_fixed = self._fix_parameter_types(value)
        min_value_fixed = self._fix_parameter_types(min_value)
        max_value_fixed = self._fix_parameter_types(max_value)

        min_value_fixed, max_value_fixed = self._fix_shapes(value_fixed, min_value_fixed, max_value_fixed)

        self._shape = value_fixed.shape
        self._length = int(np.prod(value_fixed.shape))

        self._offset = min_value_fixed
        self._scale = jnp.abs(max_value_fixed - min_value_fixed)

        # if this quantity is dependent on/calculated from other quantities
        self._dependent = False
        # all Quantities that this quantity calculates from
        self._dependencies = list()
        # the relation function that calculated value from dependencies
        self._relation = None

        # all Quantities that use this quantity to calculate value
        self._dependents = list()

        self.set_value(value)

    @staticmethod
    def _fix_shapes(value: Array, min_value: Array, max_value: Array) -> tuple[Array, Array]:
        # If value is an array and the bounds are floats, the same bounds are used for all values. The floats are
        # converted into constant arrays.
        if not np.size(value) == 1 and np.size(min_value) == 1:
            min_value = jnp.array(min_value * np.ones_like(value))
        if not np.size(value) == 1 and np.size(max_value) == 1:
            max_value = jnp.array(max_value * np.ones_like(value))

        # Values and bounds that are arrays of different length can not be handled
        if value.shape != min_value.shape or value.shape != max_value.shape:
            raise IncompatibleQuantityException("The value and the boundaries must have the same shape")

        return min_value, max_value

    @staticmethod
    def _fix_parameter_types(param: Array | float) -> Array:
        """
        Makes sure that the parameter is a jax numpy array of type jnp.float64. Primitive floats are wrapped into a
        1d-array
        """
        p = jnp.array([param]) if np.shape(param) == () else jnp.array(param)
        return p.astype(jnp.float64)

    @property
    def dependencies(self) -> list[Quantity]:
        """If dependent, get a list of parameter dependencies.

        If calculated from other quantities by a relation, this method returns
        the list of parameters that this quantity is calculated from,
        otherwise an empty list is returned.

        Returns
        -------
        list[Quantity]
            List of parameter dependencies.
        """
        if self._dependent:
            return self._dependencies
        return []

    @property
    def dependents(self) -> list[Quantity]:
        """Get a list of parameter dependents.

        This method returns the list of parameters that use this Quantity to
        calculate its value from. If no other Quantities calculate their value
        using this quantity, returns an empty list.

        Returns
        -------
        list[Quantity]
            List of parameter dependencies.
        """
        return self._dependents

    @classmethod
    def relational(
        cls,
        quantities: Quantity | list[Quantity],
        relation: Callable,
        unit: str | None = None,
        name: str | None = None,
        two_pi: bool = False,
    ) -> Self:
        """Create a relationally derived parameter.

        Creates a Quantity object that represents a Quantity that is calculated
        from other quantities using the relation function.

        Parameters
        ----------
        quantities: Quantity | list[Quantity]
            The quantities from which to calculate the value of self.
        relation: Callable
            Function describing how to calculate the value of self from other
            Quantities.
        unit: str | None
            The unit of the resulting Quantity. If 'None', then the units of all
            quantities are assumed the same.
        name: str | None
            A string identifier name of the resulting Quantity. If 'None', then
            a name is generated from the names of the related Quantities.
        two_pi: bool
            Divide by two pi for representation.

        Returns
        -------
        Quantity
            The Quantity with a relation set up, which recalculates the value of
            self from all dependencies.

        Raises
        ------
            ValueError:
                If any quantities do not have the same unit and no special
                unit is specified.

        """
        quantities = quantities if isinstance(quantities, list) else [quantities]
        min_val = jnp.min(jnp.asarray([qty.get_min_value() for qty in quantities]))
        max_val = jnp.max(jnp.asarray([qty.get_max_value() for qty in quantities]))

        if name is None:
            name = "relation_of"
            for qty in quantities:
                name += "_" + qty.get_name()

        if unit is None:
            if not all(qty.get_unit() == quantities[0].get_unit() for qty in quantities):
                raise ValueError(
                    f"All quantities in creation on {name} must have the same unit if no unit is specified."
                )
            unit = quantities[0].get_unit()

        qty = cls(
            value=min_val,
            min_value=min_val,
            max_value=max_val,
            unit=unit,
            name=name,
            two_pi=two_pi,
        )
        qty.add_relation(quantities, relation)
        return qty

    @classmethod
    def relational_copy(cls, quantity: Quantity) -> Self:
        """Create a Quantity object that is a one to one copy of a Quantity.

        If the quantity is updated, so is this relational copy.

        Parameters
        ----------
        quantity: Quantity | List[Quantity]
            The quantities from which the relational copy should be created.

        Returns
        -------
        Quantity
            The Quantity with a relation set up, which recalculates the value of
            self from all dependencies.
        """
        qty = cls(
            value=quantity.get_value(),
            min_value=quantity.get_min_value(),
            max_value=quantity.get_max_value(),
            unit=quantity.get_unit(),
            name=quantity.get_name(),
            two_pi=quantity._two_pi,
        )
        qty.add_relation(quantity, lambda x: x)
        return qty

    @property
    def dependent(self):
        """The dependency status of the quantity.

        if True:
            The value of this quantity is calculated from other quantities
        if False:
            The value of this quantity is independent of any other quantity
        """
        return self._dependent

    def add_relation(
        self,
        other: Quantity | list[Quantity],
        relation: Callable,
        check_units: bool = True,
    ) -> None:
        """Add a relation of self to one or more other quantities.

        Parameters
        ----------
        other: Quantity | List[Quantity]
            The quantities from which to calculate the value of self.
        relation: Callable
            Function describing how to calculate the value of self from other
            Quantities.
        check_units: bool
            If False, the check for equal units is not performed and unequal
            units are allowed.

        """
        other = other if isinstance(other, list) else [other]

        if not all(qty.get_unit() == self.get_unit() for qty in other) and check_units:
            raise ValueError(
                "Not all Quantities in the relation have the same units. "
                "This may lead to unintentional physical errors. "
                "Set 'check_units=False' if this behavior is wanted."
            )

        self._dependent = True
        self._dependencies = other
        self._relation = relation
        self.update()

        for qty in self._dependencies:
            qty._dependents.append(self)

    def update(self):
        """Update value of the parameter.

        Update function that is called if a value that this quantity
        is dependent on is changed.

        """
        self._set_value(self._relation(*[qty.get_value() for qty in self._dependencies]))

    def get_value(self) -> Array:
        """Get value of the parameter."""
        return self._scale * (self._value + 1) / 2 + self._offset

    def get_reduced_value(self) -> Array:
        """Return the value in the reduced representation.

        Returns
        -------
        Array
            Value from the reduced representation.

        """
        return jnp.reshape(self._value, (-1, 1))

    def set_value(self, value: Array | float) -> None:
        """Set the value of this quantity.

        Value needs to be within the range of 'min_value' and 'max_value'.

        Parameters
        ----------
        value
            Input value to be used for setting.

        Raises
        ------
        ValueError
            If the value is not within the range of 'min_value' and 'max_value', if the shape of the value is different
            from 'min_value' or 'max_value', or if this is a dependent quantity

        """
        if self._dependent:
            raise ValueError(
                "Cannot set value on dependent quantities, \
                    as it is calculated from other quantities."
            )

        self._set_value(value)

    def _set_value(self, value: Array | float) -> None:
        """Set value for the parameter."""
        if jnp.any(self._scale < float_info.epsilon):
            raise ValueError(
                f"The range between the minimum ({self._to_string(self.get_min_value())}) "
                f"and maximum ({self._to_string(self.get_max_value())}) values is too "
                f"small. Consider changing the bounds or use reduced units."
            )
        val = self._fix_parameter_types(value)
        if val.shape != self._shape:
            raise IncompatibleQuantityException("The new value must have the same shape as the old value")

        tmp = 2 * (np.reshape(val, self._shape) - self._offset) / self._scale - 1

        if jnp.any(jnp.abs(tmp) > 1.0):
            print("Error: ", val, self.get_min_value(), self.get_max_value())
            raise ValueError(
                f"Value {self._to_string(val)} out of bounds for quantity with "
                f"min_val: {self._to_string(self.get_min_value())} and "
                f"max_val: {self._to_string(self.get_max_value())}",
            )
        self._value = tmp

        # update all Quantities that depend on self
        for qty in self._dependents:
            qty.update()

    def set_reduced_value(self, value: Array | float) -> None:
        """Set reduced value limit for parameter."""
        value_fixed = self._fix_parameter_types(value)
        if value_fixed.shape != self._shape:
            raise IncompatibleQuantityException(
                f"The new value must have the same shape as the old value. \
                Got shape {value_fixed.shape}, expected shape {self._shape}."
            )
        self._value = value_fixed

    def get_min_value(self) -> Array:
        """Get minimum value of parameter."""
        return self._offset

    def get_max_value(self) -> Array:
        """Get maximum value of parameter."""
        return self._scale + self._offset

    def get_scale(self) -> Array:
        """Get scale of parameter."""
        return self._scale

    def get_length(self) -> int:
        """Get length of parameter."""
        return self._length

    def set_limits(self, min_value: Array | float, max_value: Array | float) -> None:
        """Set the allowed minimum and maximum of this quantity.

        Parameters
        ----------
        min_value : int
            Input value for setting the minimum limit.
        max_value : int
            Input value for setting the maximum limit.

        """
        old_value = self.get_value()
        min_value_fixed = self._fix_parameter_types(min_value)
        max_value_fixed = self._fix_parameter_types(max_value)

        min_value_fixed, max_value_fixed = self._fix_shapes(old_value, min_value_fixed, max_value_fixed)

        self._offset = min_value_fixed
        self._scale = np.abs(max_value_fixed - min_value_fixed)
        # the value is based on offset and scale and needs to be updated
        self._set_value(old_value)

    def set_value_and_limits(self, value: Array | float, min_value: Array | float, max_value: Array | float) -> None:
        """
        This can be used to set the value and the limits to new values at the same time. This function does not raise
        an exception if the new value is outside of the old limits.
        """
        value_fixed = self._fix_parameter_types(value)
        min_value_fixed = self._fix_parameter_types(min_value)
        max_value_fixed = self._fix_parameter_types(max_value)

        min_value_fixed, max_value_fixed = self._fix_shapes(value_fixed, min_value_fixed, max_value_fixed)

        self._offset = min_value_fixed
        self._scale = np.abs(max_value_fixed - min_value_fixed)
        self._set_value(value_fixed)

    def get_name(self) -> str:
        """Return the symbol or description or this quantity.

        Note that this does not have to be unique.
        For uniquely identifying a quantity, use getUUID.

        Returns
        -------
        str
            Value of the name attribute.

        """
        return self._name

    def set_name(self, name: str) -> None:
        """Assigns a new name to this quantity."""
        self._name = name

    def get_unit(self) -> str:
        """Get unit of measurement from parameter."""
        return self._unit

    def is_scalar(self) -> bool:
        """Check if parameter is scalar."""
        return self._length == 1

    def is_vector(self) -> bool:
        """Check if parameter is vector."""
        return self._length > 1 and len(self._shape) == 1

    # Python specific magic functions that enalble the use of e.g., binary math operators
    def __add__(self, other) -> Quantity:
        """Magic method for addition by operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(self.get_value() + other)
        return out_val

    def __radd__(self, other) -> Quantity:
        """Magic method for addition by right-hand operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(self.get_value() + other)
        return out_val

    def __sub__(self, other) -> Quantity:
        """Magic method for subtraction by operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(self.get_value() - other)
        return out_val

    def __rsub__(self, other) -> Quantity:
        """Magic method for subtraction by right-hand operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(other - self.get_value())
        return out_val

    def __mul__(self, other) -> Quantity:
        """Magic method for multiplication by operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(self.get_value() * other)
        return out_val

    def __rmul__(self, other) -> Quantity:
        """Magic method for multiplication by right-hand operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(self.get_value() * other)
        return out_val

    def __pow__(self, other) -> Quantity:
        """Magic method for exponentiation by operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(jnp.float_power(self.get_value(), other))
        return out_val

    def __rpow__(self, other) -> Quantity:
        """Magic method for exponentiation by right-hand operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(jnp.float_power(other, self.get_value()))
        return out_val

    def __truediv__(self, other) -> Quantity:
        """Magic method for division by operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(self.get_value() / other)
        return out_val

    def __rtruediv__(self, other) -> Quantity:
        """Magic method for division by right-hand operand."""
        out_val = copy.deepcopy(self)
        out_val.set_value(other / self.get_value())
        return out_val

    def __mod__(self, other) -> Quantity:
        """Magic method for representation of the modulus operation."""
        out_val = copy.deepcopy(self)
        out_val.set_value(self.get_value() % other)
        return out_val

    def __lt__(self, other) -> bool:
        """Magic method for representation of less-than operation.

        Raises
        ------
        IncompatibleQuantityException
            If the parameter is incompatible for this operation.

        Returns
        -------
        bool
            True if self's value is less than other paramter's value.
            Note: Because mypy doesn't understand what the type of
            'self.get_value' and 'other.get_value' is, the return
            type might have to be written as 'Any'.

        """
        if not self.is_scalar():
            raise IncompatibleQuantityException("Ordering operators are only usable with scalar quantities")
        return bool(self.get_value().item() < other.get_value().item())

    def __le__(self, other) -> bool:
        """Magic method for representation of less-equal operation.

        Raises
        ------
        IncompatibleQuantityException
            If the parameter is incompatible for this operation.

        Returns
        -------
        bool
            True if self's value is less than or equal to the operand's
            value.
            Note: Because mypy doesn't understand what the type of
            'self.get_value' and 'other.get_value' is, the return
            type might have to be written as 'Any'.

        """
        if not self.is_scalar():
            raise IncompatibleQuantityException("Ordering operators are only usable with scalar quantities")
        return bool(self.get_value().item() <= other.get_value().item())

    def __eq__(self, other) -> bool:
        """Magic method for representation of equality operation."""
        if self._shape != other._shape:
            return False
        return all(self.get_value() == other.get_value())

    def __ne__(self, other) -> bool:
        """Magic method for representation of not-equal operation."""
        if self._shape != other._shape:
            return True
        return any(self.get_value() != other.get_value())

    def __ge__(self, other) -> bool:
        """Magic method for representation of greater-equal operation.

        Raises
        ------
        IncompatibleQuantityException
            If the parameter is incompatible for this operation.

        Returns
        -------
        bool
            True if self's value is greater than or equal to the operand's
            value.
            Note: Because mypy doesn't understand what the type of
            'self.get_value' and 'other.get_value' is, the return
            type might have to be written as 'Any'.

        """
        if not self.is_scalar():
            raise IncompatibleQuantityException("Ordering operators are only usable with scalar quantities")
        return bool(self.get_value().item() >= other.get_value().item())

    def __gt__(self, other) -> bool:
        """Magic method for representation of greater-than operation.

        Raises
        ------
        IncompatibleQuantityException
            If the parameter is incompatible for this operation.

        Returns
        -------
        bool
            True if self's value is greater than the operand's value.
            Note: Because mypy doesn't understand what the type of
            'self.get_value' and 'other.get_value' is, the return
            type might have to be written as 'Any'.

        """
        if not self.is_scalar():
            raise IncompatibleQuantityException("Ordering operators are only usable with scalar quantities")
        return bool(self.get_value().item() > other.get_value().item())

    def __array__(self):
        """Magic method for representation into array."""
        return np.array(self.get_value())

    def __jax_array__(self):
        """Magic method for representation into array."""
        return jnp.array(self.get_value())

    def __len__(self):
        """Magic method for calculation of length."""
        return self._length

    def __getitem__(self, key):
        """Magic method for selection of item.

        Parameters
        ----------
        key: int
            Index of object for retrieval.

        """
        if self._length == 1 and key == 0:
            return self.get_value()
        return self.get_value().__getitem__(key)

    def __abs__(self):
        """Magic method for absolute value calculation."""
        return abs(self.get_value())

    def __float__(self):
        """Magic method for float conversion.

        Raises
        ------
        NotImplementedError
            If the length of the parameter is greater than 1.

        """
        if self._length > 1:
            raise NotImplementedError
        return float(np.squeeze(self.get_value()))

    def __repr__(self):
        """Magic method for human readable representation."""
        return self.__str__()

    def __str__(self):
        """Human readable representation of the parameters set to optimize."""
        return self._to_string(self.get_value())

    def _to_string(self, val: Array):
        """Represent parameter as custom defined string value."""
        ret = ""
        if len(val) > 1:
            ret = "["
            delimiter = ", "
        else:
            delimiter = " "
        for ii, entry in enumerate(val):
            if ii == len(val) - 1:
                delimiter = ""
            if self._unit != "":
                if self._two_pi:
                    ret += Quantity._make_human_readable(entry / np.pi / 2) + self._unit + " x 2pi" + delimiter
                else:
                    ret += Quantity._make_human_readable(entry) + self._unit + delimiter
            else:
                if self._two_pi:
                    ret += Quantity._make_human_readable(entry / np.pi / 2, use_prefix=False) + " x 2pi" + delimiter
                else:
                    ret += Quantity._make_human_readable(entry, use_prefix=False) + delimiter
        if self._name:
            if len(val) > 1:
                ret = self._name + ": " + ret + "]"
            else:
                ret = self._name + ": " + ret
        else:
            if len(val) > 1:
                ret = ret + "]"
        return ret

    @staticmethod
    def _make_human_readable(val, use_prefix: bool = True) -> str:
        """Convert to human readable string in engineering notation.

        Parameters
        ----------
        use_prefix: bool=True
            Adds a prefix string derived from '_engineering_number'
            to the final format string.

        Returns
        -------
        str
            Human readable formatted string for value.

        """
        if use_prefix:
            num, prefix = Quantity._engineering_number(val)
            formatted_string = f"{num:.3g} " + prefix
        else:
            formatted_string = f"{val:.3g}"
        return formatted_string

    # Internal utility functions
    @staticmethod
    def _engineering_number(val: float) -> tuple[float, str]:
        """Convert number to engineering notation.

        Returns number and prefix.

        Parameters
        ----------
        val: float
            Input number to be converted to engineering notation.

        Returns
        -------
        tuple[float, str]
            Engineering notation composite made of the number and the prefix.

        """
        if np.isnan(val):
            return np.nan, "NaN"

        sign = 1.0
        if val == 0:
            return 0.0, ""
        if val < 0.0:
            val = -val
            sign = -1.0
        tmp = np.log10(val)
        idx = int(tmp // 3)

        if tmp < 0:
            units = ["m", "µ", "n", "p", "f", "a", "z"]
            if np.abs(idx) > len(units):
                return val * (10 ** (3 * len(units))), units[-1]
            prefix = units[-(idx + 1)]
        else:
            units = ["", "K", "M", "G", "T", "P", "E", "Z"]
            if np.abs(idx) > len(units) - 1:
                return val * (10 ** (-3 * (len(units) - 1))), units[-1]
            prefix = units[idx]

        return sign * (10 ** (tmp % 3)), prefix

    def to_dict(self) -> dict:
        """
        Creates a dictionary representation of this quantity that can be stored. The returned dict is compatible with
        the from_dict function, i.e. the quantity can be fully restored including its bounds, name, unit, etc. Higher
        dimensional quantities (tensors) will be flattened into a list but their proper shape is stored as well.
        """
        if self.dependent:
            raise UserWarning("Saving of dependent quantities is not supported yet")

        return {
            "unit": self._unit,
            "shape": self._shape,
            "two_pi": self._two_pi,
            "value": self.get_value().flatten().tolist(),
            "min": self.get_min_value().tolist(),
            "max": self.get_max_value().tolist(),
        }

    def from_dict(self, data: dict) -> None:
        """
        Loads the quantity from a dictionary. The dictionary must have the same form as the one created by the toDict
        function. All properties of this quantity (value, name, etc.) will be overwritten.
        """
        self._unit = data["unit"]
        self._shape = data["shape"]
        self._length = int(np.prod(self._shape))
        self._two_pi = data["two_pi"]

        # The value and limits need to be set at the same time so that the new value is not out of range
        value = np.array(data["value"]).reshape(self._shape)
        min_val = np.array(data["min"]).reshape(self._shape)
        max_val = np.array(data["max"]).reshape(self._shape)
        self.set_value_and_limits(value, min_val, max_val)



"""Test composite optimization."""

import numpy as np
import pytest

from paraqeet.measurement.unitary_fidelity import UnitaryFidelity
from paraqeet.model.closed_system import ClosedSystem
from paraqeet.model.composite_hamiltonian import CompositeHamiltonian
from paraqeet.model.coupling import TwoBodyCoupling
from paraqeet.model.drive_operator import DriveOperator
from paraqeet.model.transmon import Transmon
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.scipy_optimizer import ScipyOptimizer
from paraqeet.optimizers.scipy_optimizer_gradient import ScipyOptimizerGradient
from paraqeet.propagation.scipy_expm_goat import ScipyExpmGOAT
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import FlatTopGaussianEnvelope
from paraqeet.signal.iq_mixer import IQMixer

FREQ1 = 5.5e9 * 2 * np.pi
ANHARM1 = -240e6 * 2 * np.pi

FREQ2 = 6.0e9 * 2 * np.pi
ANHARM2 = -200e6 * 2 * np.pi

COUPLINGSTR = 25e6 * 2 * np.pi

T_FINAL = 150e-9
RES = 100e9


@pytest.fixture
def tone():
    """Create a signal tone."""

    def _method(amp, t_final):
        tone = FlatTopGaussianEnvelope(
            amplitude=Quantity(
                amp * 2 * np.pi,
                min_value=0.8 * amp * 2 * np.pi,
                max_value=1.2 * amp * 2 * np.pi,
                unit="Hz",
            ),
            t_final=Quantity(
                t_final,
                min_value=0.8 * t_final,
                max_value=1.2 * t_final,
                unit="Hz",
            ),
        )
        return tone

    return _method


@pytest.fixture
def coupled_transmons(tone):
    """Create a coupled Transmon system."""
    freq1 = 6.0e9
    freq2 = 6.0002e9

    tone1 = tone(191e6, T_FINAL)
    tone2 = tone(9.18e6, T_FINAL)

    generator1 = IQMixer(
        envelopes=[tone1],
        frequency=Quantity(
            freq1 * 2 * np.pi,
            min_value=np.array(0.8 * freq1 * 2 * np.pi),
            max_value=np.array(1.2 * freq1 * 2 * np.pi),
            unit="Hz",
        ),
        phase=Quantity(
            0.01,
            min_value=np.array(-np.pi),
            max_value=np.array(np.pi),
            unit="rad",
        ),
    )
    drive1 = DriveOperator(generator1, is_longitudinal=False)

    generator2 = IQMixer(
        envelopes=[tone2],
        frequency=Quantity(
            freq2 * 2 * np.pi,
            min_value=np.array(0.8 * freq2 * 2 * np.pi),
            max_value=np.array(1.2 * freq2 * 2 * np.pi),
            unit="Hz",
        ),
        phase=Quantity(
            -0.39720756,
            min_value=np.array(-np.pi),
            max_value=np.array(np.pi),
            unit="rad",
        ),
    )
    drive2 = DriveOperator(generator2, is_longitudinal=False)

    transmon1 = Transmon(
        dimension=3,
        frequency=Quantity(FREQ1, np.array(0.8 * FREQ1), np.array(1.2 * FREQ1), "Hz"),
        anharmonicity=Quantity(ANHARM1, np.array(1.2 * ANHARM1), np.array(0.8 * ANHARM1), "Hz"),
        drives=[drive1],
    )
    transmon2 = Transmon(
        dimension=3,
        frequency=Quantity(FREQ2, np.array(0.8 * FREQ2), np.array(1.2 * FREQ2), "Hz"),
        anharmonicity=Quantity(ANHARM2, np.array(1.2 * ANHARM2), np.array(0.8 * ANHARM2), "Hz"),
        drives=[drive2],
    )
    coupling = TwoBodyCoupling(
        transmon1,
        transmon2,
        is_longitudinal=False,
        coefficient=Quantity(
            COUPLINGSTR,
            np.array(0.8 * COUPLINGSTR),
            np.array(1.2 * COUPLINGSTR),
            "Hz",
        ),
    )
    hamiltonian = CompositeHamiltonian([transmon1, transmon2], [coupling])
    model = ClosedSystem(hamiltonian)
    prop = ScipyExpmGOAT(model=model, resolution=RES)

    pauli_x = np.array([[0.0, 1], [1, 0.0]])
    pauli_z = np.array([[1, 0], [0.0, -1]])
    pauli_zx = np.exp(1j * np.pi / 4) * np.kron(pauli_z, pauli_x)
    cr_gate = np.array([[1.0, 0, 0, 0], [0, 1.0, 0, 0], [0, 0, 0, 1.0], [0, 0, 1.0, 0]])

    cr_gate = pauli_zx @ cr_gate
    prop.set_initial_state(np.identity(9))
    gate_fid = UnitaryFidelity(
        propagation=prop,
        gate=cr_gate,
    )

    tone1_amp = tone1.get_parameters()[0]

    optmap = OptimizationMap()
    optmap.add(tone1, [tone1_amp])
    return gate_fid, optmap


@pytest.fixture
def opt(coupled_transmons):
    """Return Scipy optimizer from coupled transmons."""
    measure, optmap = coupled_transmons
    opt = ScipyOptimizer(measure, optimization_map=optmap)
    opt.set_options({"maxiter": 5})
    return opt


@pytest.fixture
def grad_opt(coupled_transmons):
    """Return Scipy optimizer gradient."""
    measure, optmap = coupled_transmons
    opt = ScipyOptimizerGradient(measure, optimization_map=optmap)
    opt.set_options({"maxiter": 2})
    return opt


@pytest.mark.skip(reason="Projection needed")
def test_optim_finite_diff(opt):
    """Test optimization via finite differences."""
    # TODO: Add projection before testing.
    res = opt.optimize(times=T_FINAL)
    assert res.value < 0.1


@pytest.mark.skip(reason="Projection needed")
def test_optim_goat(grad_opt):
    """Test GOAT optimization."""
    # TODO: Add projection before testing.
    res = grad_opt.optimize(times=T_FINAL)
    assert res.value < 0.1
